###############################################################################
# Copyright (c) 2002-2009, Impulse Accelerated Technologies, Inc.
# All Rights Reserved.
#
# Generate bus interfaces for Xilinx platforms
#
# These procedures must be called from within impulse_arch.exe.
#
###############################################################################

######################################################
# UTILITY PROCS
######################################################

proc lmember {l m} {
  return [expr [lsearch $l $m] >= 0]
}

proc makeVec {val width} {
	set spec "%0[expr ($width + 3) / 4]x"
	set hex [format $spec $val]
	set val ""
	foreach digit [split $hex ""] {
	    switch $digit {
		0 { append val "0000" }
		1 { append val "0001" }
		2 { append val "0010" }
		3 { append val "0011" }
		4 { append val "0100" }
		5 { append val "0101" }
		6 { append val "0110" }
		7 { append val "0111" }
		8 { append val "1000" }
		9 { append val "1001" }
		a { append val "1010" }
		b { append val "1011" }
		c { append val "1100" }
		d { append val "1101" }
		e { append val "1110" }
		f { append val "1111" }
	    }
	}
	set val [string range $val [expr [string length $val] - $width] end]
	return \"$val\"
}

proc checkInstruction {con} {
    set ret [lindex $con 4]
    if {$ret != ""} {
	error "Instruction cannot return a value; use a pointer parameter."

    }
    set inputs 0
    set outputs 0
    set args [lindex $con 3]
    foreach arg $args {
	set name [lindex $arg 0]
	if {[lindex $arg 2] == "input"} {
	    if {[string compare -nocase $name "wildcard"] != 0} {
		if {[lindex $arg 1] != 32} {
		    error "UDI input $name must be 32 bits."
		}
		incr inputs
	    } else {
		if {[lindex $arg 1] != 5} {
		    error "UDI wildcard must be 5 bits."
		}
	    }
	} else {
	    if {[string compare -nocase [string range $name 0 1] "cr"] != 0} {
		if {[lindex $arg 1] != 32} {
		    error "UDI result $name must be 32 bits."
		}
		incr outputs
	    } elseif {![string is integer [string range $name 2 end]]} {
		if {[lindex $arg 1] != 32} {
		    error "UDI result $name must be 32 bits."
		}
		incr outputs
	    } else {
		if {[lindex $arg 1] != 4} {
		    error "UDI $name must be 4 bits."
		}
	    }
	}
    }
    if {$inputs > 2} {
	error "Instruction cannot have more than 2 input parameters."
    }
    if {$outputs > 1} {
	error "Instruction cannot have more than 1 output parameter."
    }
}

proc instructionHasResult {con} {
    set args [lindex $con 3]
    foreach arg $args {
	if {[lindex $arg 2] != "input"} {
	    return 1
	}
    }
    return 0
}

proc instructionReturn {con} {
    set args [lindex $con 3]
    foreach arg $args {
	if {[lindex $arg 2] != "input"} {
	    set name [lindex $arg 0]
	    if {[string compare -nocase [string range $name 0 1] "cr"] != 0} {
		return $name
	    } elseif {![string is integer [string range $name 2 end]]} {
		return $name
	    }
	}
    }
    return ""
}

proc instructionCR {con} {
    set args [lindex $con 3]
    foreach arg $args {
	if {[lindex $arg 2] != "input"} {
	    set name [lindex $arg 0]
	    if {[string compare -nocase [string range $name 0 1] "cr"] == 0} {
		if {[string is integer [string range $name 2 end]]} {
		    return $name
		}
	    }
	}
    }
    return ""
}

# FSL and DVI only support streams.  Substitutes another bus for
#   any non-stream connection to FSL.
# Also chooses the "generic" bus for connections to co_ports, if they
#   are not to be connected to a real bus.
proc selectBus {con} {
	global _flags  ;# Set by impulse_arch.exe

	set type [lindex $con 0]
	if {$type == "memory"} {
		set busname [lindex $con 2]
		set connectedToType [lindex $con 3]
	} elseif {$type == "instruction"} {
		set busname [lindex $con 2]
		set connectedToType hp
	} else {
		set busname [lindex $con 4]
		set connectedToType [lindex $con 5]
	}

        if {$_flags(no_port_bus_connect) && $connectedToType == "port"} {
	        set busname "port"
	} elseif {$busname == ""} {
		set busname "generic"
	} elseif {$busname == "fsl" && $type != "stream"} {
		set busname "opb"
	} elseif {$busname == "dvi" && $type != "stream"} {
		set busname "generic"
	} elseif {$busname == "apu" && $type != "stream" && $type != "instruction"} {
		set busname "plb"
	} elseif {$busname == "apu440" && $type != "stream" && $type != "instruction"} {
		set busname "plb4"
	} elseif {$busname == "bus_signals"} {
		set busname "port"
	}
	return $busname
}


######################################################
# PORT-SPECIFIC PROCS
######################################################

proc genPortPorts {fd con} {
	set first 1
	foreach sig [getConnectionSignals $con] {
		if {$first} {
			set first 0
		} else {
			puts $fd ";"
		}
		set name [lindex $sig 0]
		set dir [lindex $sig 1]
		set type [lindex $sig 2]
		puts -nonewline $fd "    Port_$name : $dir $type"
	}
}

proc genPortConn {fd con} {
	foreach sig [getConnectionSignals $con] {
		set name [lindex $sig 0]
		set dir [lindex $sig 1]
		set type [lindex $sig 2]
		set pname $name
		if { [string match "*_data" $name] } {
			regsub "_data$" $name "_idata" sname
		} else {
			set sname $name
		}
		if {$dir == "out"} {
			puts $fd "  Port_$pname <= $sname;"
		} else {
			puts $fd "  $sname <= Port_$pname;"
		}
	}
	puts $fd ""
}

proc genPortMPD {fd connections} {
	foreach con $connections {
		set bus [lindex $con 4]
		set co_type [lindex $con 0]
		set con_dir [lindex $con 1]
		set con_name [string toupper [lindex $con 2]]
		if {[selectBus $con] == "port"} {
			foreach sig [getConnectionSignals $con] {
				set name [lindex $sig 0]
				set dir [lindex $sig 1]
				set type [lindex $sig 2]
				
				#puts "genPortMPD:name = $name $sig"
				if {$bus == "bus_signals"} {
					set init_val ""
					set reg_result [regexp {.*_(.*$)} $name match suffix]
					if { $co_type == "stream" } {
						# keep streams consistent, _en=output, _rdy=input, connecting
						#  them to a bus criss-crosses them on the input side
						if {$con_dir=="in" } {
							if { $suffix=="rdy" } { 
								set suffix "en"
							} elseif { $suffix=="en" } {
								set suffix "rdy"
							} else {
								# do nothing
							}
						}
					} elseif { $co_type == "register" } {
						if {$con_dir=="in" } {
							if { $suffix=="en" } {
								set init_val ", INITIALVAL=VCC"
							} 
						} else {
							# ouput has '_value while input has '_data
							if { $suffix=="value" } {
								set suffix "data"
							}
						}
					} else {
						# do nothing
					}
					set def_con "co_${co_type}_${suffix}"
					set def_con_bus "${init_val}, BUS = ${con_name}"
					#puts "   def_con = $def_con"
					#puts "   reg_result = $reg_result"
					#puts "   match = $match"
					#puts "   suffix = $suffix"
				} else {
					set def_con ""
					set def_con_bus ""
				}
				
				if {$dir == "out"} {
					puts -nonewline $fd "PORT Port_$name = \"${def_con}\", DIR = O${def_con_bus}"
				} else {
					puts -nonewline $fd "PORT Port_$name = \"${def_con}\", DIR = I${def_con_bus}"
				}
				if {[string first vector $type] >= 0} {
					set start [expr [string first "(" $type] + 1]
					set end [expr [string first ")" $type] - 1]
					set range [string range $type $start $end]
					puts $fd ", VEC = \[[lindex $range 0]:[lindex $range 2]\]"
				} else {
					puts $fd ""
				}
			}
		}
	}
	puts $fd ""
}

######################################################
# APU (PPC440) - SPECIFIC PROCS
######################################################

proc genApu440Init {} {
    global apu440

    set apu440(nonautonDecodeNo) 0
    set apu440(autonDecodeNo) 0
}

proc genApu440Generics {fd multibus hasSlave hasMaster} {
#    puts $fd "    C_PCMCIA_AWIDTH    : integer                   := 42;"
#    puts $fd "    C_PCMCIA_DWIDTH    : integer                   := 16;"
    
    if {! $multibus} {
#        puts $fd "    C_BASEADDR     : std_logic_vector(0 to C_PCMCIA_AWIDTH-1) := X\"3FFF0000000\""
    } else {
#        puts $fd "    C_PCMCIA_BASEADDR  : std_logic_vector(0 to C_PCMCIA_AWIDTH-1) := X\"3FFF0000000\""
    }
}

proc genApu440Ports {fd hasSlave hasMaster} {
    puts $fd "    apufcminstruction : std_logic_vector(0 to 31);"
    puts $fd "    apufcminstrvalid : std_logic;"
    puts $fd "    apufcmradata : std_logic_vector(0 to 31);"
    puts $fd "    apufcmrbdata : std_logic_vector(0 to 31);"
    puts $fd "    apufcmoperandvalid : std_logic;"
    puts $fd "    apufcmflush : std_logic;"
    puts $fd "    apufcmwritebackok : std_logic;"
    puts $fd "    apufcmloaddata : std_logic_vector(0 to 127);"
    puts $fd "    apufcmloaddvalid : std_logic;"
    puts $fd "    apufcmloadbyteaddr : std_logic_vector(0 to 3);"
    puts $fd "    apufcmendian : std_logic;"
    puts $fd "    apufcmdecudi : std_logic_vector(0 to 3);"
    puts $fd "    apufcmdecudivalid : std_logic;"
    puts $fd "    apufcmdecfpuop : std_logic;"
    puts $fd "    apufcmdecload : std_logic;"
    puts $fd "    apufcmdecstore : std_logic;"
    puts $fd "    apufcmdecldstxfersize : std_logic_vector(0 to 2);"
    puts $fd "    apufcmdecnonauton : std_logic;"
    puts $fd "    apufcmnextinstrready : std_logic;"
    puts $fd "    fcmapuresult : inout std_logic_vector(0 to 31);"
    puts $fd "    fcmapudone : inout std_logic;"
    puts $fd "    fcmapusleepnotready : inout std_logic;"
    puts $fd "    fcmapuresultvalid : inout std_logic;"
    puts $fd "    fcmapucr : inout std_logic_vector(0 to 3);"
    puts $fd "    fcmapuexception : inout std_logic;"
    puts $fd "    fcmapustoredata : inout std_logic_vector(0 to 127);"
    puts $fd "    fcmapuconfirminstr : inout std_logic;"
    puts $fd "    fcmapufpscrfex : inout std_logic;"
    puts $fd "udidebug : out std_logic_vector(153 downto 0);"
    puts -nonewline $fd "    apu_clk, reset : in std_logic"
}

proc genApu440ConnSignals {fd con} {
    set type [lindex $con 0]
    if {$type != "memory" && $type != "instruction"} {
        set iname [lindex $con 2]
        puts $fd "  signal ${iname}_storedata : std_logic_vector(0 to 127);"
        puts $fd "  signal ${iname}_done : std_logic;"
        puts $fd "  signal ${iname}_sleepnotready : std_logic;"
        puts $fd "  signal ${iname}_resultvalid : std_logic;"
        puts $fd "  signal ${iname}_cs : std_logic;"
        puts $fd "  signal ${iname}_instrvalid : std_logic;"
    } elseif {$type == "instruction"} {
	set iname [lindex $con 1]
        puts $fd "  signal ${iname}_done : std_logic;"
        puts $fd "  signal ${iname}_sleepnotready : std_logic;"
	if {[instructionHasResult $con]} {
	    puts $fd "  signal ${iname}_resultvalid : std_logic;"
	}
    }
}

proc genApu440Signals {fd noAttachments} {
    set selWidth [calcAddrBits $noAttachments]
    set addrWidth [expr $selWidth + 4]

    if {$selWidth > 0} {
	puts $fd "  signal apusel : std_logic_vector(0 to [expr $selWidth - 1]);"
    }
    puts $fd "  signal nonautonomousWildcardDecodeValid : std_logic;"
    puts $fd "  signal autonomousWildcardDecodeValid : std_logic;"
    puts $fd "  signal wildcardDecode : std_logic_vector(0 to 3);"
    puts $fd "  signal fcm_transaction_id : unsigned (31 downto 0);"
}

proc genApu440Ipif {fd multibus noAttachments} {
    set selWidth [calcAddrBits $noAttachments]
    set firstSelBit [expr 9 - $selWidth]

    puts $fd {
  udidebug(0) <= apu_clk;
  udidebug(1) <= fcmapuresultvalid;
  udidebug(2) <= fcmapudone;
  udidebug(3) <= fcmapusleepnotready;
  udidebug(4) <= fcmapuexception;
  udidebug(5) <= fcmapuconfirminstr;

  udidebug(37 downto 6) <= apufcminstruction;
  udidebug(38) <= apufcmdecudivalid;
  udidebug(39) <= apufcmflush;
  udidebug(40) <= apufcmwritebackok;
  udidebug(41) <= apufcminstrvalid;
  udidebug(42) <= apufcmoperandvalid;
  udidebug(43) <= apufcmdecnonauton;
  udidebug(44) <= apufcmdecstore;
  udidebug(45) <= apufcmdecload;
  udidebug(46) <= apufcmloaddvalid;
  udidebug(49 downto 47) <= apufcmdecldstxfersize;
  udidebug(53 downto 50) <= apufcmdecudi;
  udidebug(85 downto 54) <= apufcmradata;
  udidebug(117 downto 86) <= fcmapuresult;
  udidebug(121 downto 118) <= apufcmloadbyteaddr;
  udidebug(153 downto 122) <= std_logic_vector(fcm_transaction_id);

  process (reset, apu_clk)
  begin
    if (reset = '1') then
	fcm_transaction_id <= X"00000000";
    elsif (apu_clk'event and apu_clk='1') then
        if (apufcminstrvalid = '1' and fcmapudone = '1') then
	  fcm_transaction_id <= fcm_transaction_id + X"00000001";
        end if;
    end if;
  end process;

}

    # bits 6-10 of the instruction are the "RT" field, and bits 9,10 are internal to a stream.
    # this leaves stream-selection to bits 6-8 and thus limits us to 8 streams.
    puts $fd "  apusel <= apufcminstruction($firstSelBit to 8);\n"
    puts $fd "  nonautonomousWildcardDecodeValid <= '1' when apufcmdecudivalid = '1' and apufcmdecudi = \"0000\" else '0';"
    puts $fd "  autonomousWildcardDecodeValid <= '1' when apufcmdecudivalid = '1' and apufcmdecudi = \"0001\" else '0';"
    puts $fd "  wildcardDecode <= apufcminstruction(22 to 25);"
    puts $fd ""
}

# generates code to OR together the output signals from each stream.
# also ties all the unused fcm->apu lines to ground.
proc genApu440OutputSignals {fd cons} {
#    puts $fd "cons has $[llength $cons] ([lindex $cons 0])"

    foreach output [list storedata done sleepnotready resultvalid] {
        puts -nonewline $fd "  fcmapu${output} <= "
        set iter 0
        foreach con $cons {
            set type [lindex $con 0]
				if {$type != "instruction"} {
					if {$type == "stream" && [selectBus $con] == [lindex $con 4] } {  # only streams are currently supportted
						set dir [lindex $con 1]
						set iname [lindex $con 2]
						if {$iter > 0} {
							puts -nonewline $fd " or ";
						}
						puts -nonewline $fd "${iname}_${output}";
						incr iter
					}
				} elseif {$output != "storedata"} {
					if {$output != "resultvalid" || [instructionHasResult $con]} {
						set iname [lindex $con 1]
						if {$iter > 0} {
							puts -nonewline $fd " or ";
						}
						puts -nonewline $fd "${iname}_${output}";
						incr iter
					}
				}
        }
	if {$iter == 0} {
	    if {$output == "storedata"} {
		puts -nonewline $fd "(others => '0')"
	    } else {
		puts -nonewline $fd "'0'"
	    }
	}
        puts $fd ";"
    }

    puts -nonewline $fd "  fcmapuresult <= "
    set iter 0
    foreach con $cons {
	set type [lindex $con 0]
	if {$type == "instruction"} {
	    if {[instructionReturn $con] != ""} {
		set iname [lindex $con 1]
		if {$iter > 0} {
		    puts -nonewline $fd " or ";
		}
		puts -nonewline $fd "std_logic_vector(p_${iname}_[instructionReturn $con])";
		incr iter
	    }
	}
    }
    if {$iter == 0} {
	puts -nonewline $fd "X\"00000000\""
    }
    puts $fd ";"

    puts -nonewline $fd "  fcmapucr <= "
    set iter 0
    foreach con $cons {
	set type [lindex $con 0]
	if {$type == "instruction"} {
	    if {[instructionCR $con] != ""} {
		set iname [lindex $con 1]
		if {$iter > 0} {
		    puts -nonewline $fd " or ";
		}
		puts -nonewline $fd "std_logic_vector(p_${iname}_[instructionCR $con])";
		incr iter
	    }
	}
    }
    if {$iter == 0} {
	puts -nonewline $fd "\"0000\""
    }
    puts $fd ";"

    puts $fd "\n  -- unused signals"
    puts $fd "  fcmapuexception <= '0';"
    puts $fd "  fcmapuconfirminstr <= '1';"
    puts $fd "  fcmapufpscrfex <= '0';"
    puts $fd ""
}

proc genApu440Conn {fd con no total rstName clkName} {
    global apu440

    set selWidth [calcAddrBits $total]
    
    set type [lindex $con 0]
    if {$type != "instruction"} {
	set dir [lindex $con 1]
	set iname [lindex $con 2]
	set width [lindex $con 3]
    }
    switch $type {
        {stream} {
	    set addr [makeVec $no $selWidth]
 	        puts $fd "  ${iname}_cs <= '1' when apusel = $addr else '0';"
	    puts $fd "  ${iname}_instrvalid <= '1' when (apufcminstrvalid = '1') and (${iname}_cs = '1') and apufcmdecfpuop = '0' else '0';"
	    if {$dir == "in"} {
	        puts $fd "  ${iname}_if: apu_to_stream"
	    } else {
	        puts $fd "  ${iname}_if: stream_to_apu"
	    }
	    puts $fd "    generic map ("
	    puts $fd "      datawidth => $width"
	    puts $fd "    )"
            puts $fd "    port map ("
            puts $fd "      reset, apu_clk,"
            puts $fd "      ${iname}_cs,"
            puts $fd "      apufcminstruction,"
            puts $fd "      ${iname}_instrvalid,"
            puts $fd "      apufcmradata,"
            puts $fd "      apufcmrbdata,"
            puts $fd "      apufcmoperandvalid,"
            puts $fd "      apufcmflush,"
            puts $fd "      apufcmwritebackok,"
            puts $fd "      apufcmloaddata,"
            puts $fd "      apufcmloaddvalid,"
            puts $fd "      apufcmloadbyteaddr,"
            puts $fd "      apufcmendian,"
            puts $fd "      apufcmdecudi,"
            puts $fd "      apufcmdecudivalid,"
            puts $fd "      ${iname}_storedata,"
            puts $fd "      ${iname}_done,"
            puts $fd "      ${iname}_sleepnotready,"
            puts $fd "      ${iname}_resultvalid,"
            puts $fd "      ${iname}_en,"
            puts $fd "      ${iname}_rdy,"
            puts $fd "      ${iname}_eos,"
            puts $fd "      ${iname}_idata"
            puts $fd "    );\n"
        }
        {signal} {
            error "TODO"
        }
        {register} {
            error "TODO"
	}
        {memory} {
	    error "TODO"
        }
	{instruction} {
	    set iname [lindex $con 1]

	    set args [lindex $con 3]
	    set argno 0
	    foreach arg $args {
		if {[lindex $arg 2] == "input"} {
		    if {[string compare -nocase [lindex $arg 0] "wildcard"] != 0} {
			if {$argno == 0} {
			    puts $fd "  p_${iname}_[lindex $arg 0] <= std_ulogic_vector(apufcmradata);"
			} else {
			    puts $fd "  p_${iname}_[lindex $arg 0] <= std_ulogic_vector(apufcmrbdata);"
			}
			incr argno
		    } else {
			puts $fd "  p_${iname}_wildcard <= std_ulogic_vector(apufcminstruction(21 to 25));"
		    }
		}
	    }

	    set sideEffect [lindex $con 5]
	    set hasResult [instructionHasResult $con]
	    if {!$hasResult} {
		puts $fd "  ${iname}_if: autonomous_udi_ctrl"
	    } else {
		if {$sideEffect} {
		    puts $fd "  ${iname}_if: blocking_udi_ctrl"
		} else {
		    puts $fd "  ${iname}_if: nonblocking_udi_ctrl"
		}
	    }
	    puts $fd "    generic map ("
	    if {[lindex $con 6] == "wildcard"} {
		if {$hasResult} {
		    puts $fd "      UDI_NUM => X\"[format {%X} $apu440(nonautonDecodeNo)]\","
		    incr apu440(nonautonDecodeNo)
		} else {
		    puts $fd "      UDI_NUM => X\"[format {%X} $apu440(autonDecodeNo)]\","
		    incr apu440(autonDecodeNo)
		}
	    } else {
		puts $fd "      UDI_NUM => X\"[format {%X} $no]\","
	    }
	    puts $fd "      NUM_OPERANDS => $argno"
	    puts $fd "    )"
	    puts $fd "    port map (reset, apu_clk,"
	    if {[lindex $con 6] == "wildcard"} {
		if {$hasResult} {
		    puts $fd "      nonautonomousWildcardDecodeValid,"
		} else {
		    puts $fd "      autonomousWildcardDecodeValid,"
		}
		puts $fd "      wildcardDecode,"
	    } else {
		puts $fd "      apufcmdecudivalid,"
		puts $fd "      apufcmdecudi,"
	    }
	    puts $fd "      apufcmoperandvalid,"
	    puts $fd "      apufcmflush,"
	    puts $fd "      apufcmwritebackok,"
	    puts $fd "      p_${iname}_ack,"
	    if {$hasResult} {
		puts $fd "      ${iname}_resultvalid,"
	    }
	    puts $fd "      ${iname}_done,"
	    puts $fd "      ${iname}_sleepnotready,"
	    puts $fd "      p_${iname}_req"
	    puts $fd "    );\n"
	}
    }
}

proc genApu440MPD {fd} {
    puts $fd "OPTION IMP_NETLIST = TRUE"
    puts $fd "OPTION SIM_MODELS = BEHAVIORAL"
    puts $fd "OPTION IP_GROUP = MICROBLAZE:PPC:USER"
    
    puts $fd "OPTION RUN_NGCBUILD = TRUE"

    puts $fd "BUS_INTERFACE BUS = SFCB2, BUS_STD = FCB2, BUS_TYPE = SLAVE"

    puts $fd "PORT FCMAPURESULT = Sl_RESULT, DIR = OUT, VEC = \[0:31\], BUS = SFCB2"
    puts $fd "PORT FCMAPUDONE = Sl_DONE,DIR = OUT, BUS = SFCB2" 
    puts $fd "PORT FCMAPUSLEEPNOTREADY = Sl_SLEEPNOTREADY, DIR = OUT, BUS = SFCB2" 
    puts $fd "PORT FCMAPURESULTVALID = Sl_RESULTVALID, DIR = OUT, BUS = SFCB2" 
    puts $fd "PORT FCMAPUCR = Sl_CR, DIR = OUT, VEC = \[0:3\], BUS = SFCB2"
    puts $fd "PORT FCMAPUEXCEPTION = Sl_EXCEPTION, DIR = OUT, BUS = SFCB2" 
    puts $fd "PORT FCMAPUSTOREDATA = Sl_STOREDATA, DIR = OUT, VEC = \[0:127\], BUS = SFCB2"
    puts $fd "PORT FCMAPUCONFIRMINSTR = Sl_CONFIRMINSTR, DIR = OUT, BUS = SFCB2"
    puts $fd "PORT FCMAPUFPSCRFEX = Sl_FPSCRFEX, DIR = OUT, BUS = SFCB2"
    puts $fd "PORT APUFCMINSTRUCTION = FCB_INSTRUCTION, DIR = IN, VEC = \[0:31\], BUS = SFCB2"
    puts $fd "PORT APUFCMINSTRVALID = FCB_INSTRVALID, DIR = IN, BUS = SFCB2" 
    puts $fd "PORT APUFCMRADATA = FCB_RADATA, DIR = IN, VEC = \[0:31\], BUS = SFCB2"
    puts $fd "PORT APUFCMRBDATA = FCB_RBDATA, DIR = IN, VEC = \[0:31\], BUS = SFCB2"
    puts $fd "PORT APUFCMOPERANDVALID = FCB_OPERANDVALID,  DIR = IN, BUS = SFCB2" 
    puts $fd "PORT APUFCMFLUSH = FCB_FLUSH, DIR = IN, BUS = SFCB2" 
    puts $fd "PORT APUFCMWRITEBACKOK = FCB_WRITEBACKOK, DIR = IN, BUS = SFCB2" 
    puts $fd "PORT APUFCMLOADDATA = FCB_LOADDATA, DIR = IN, VEC = \[0:127\], BUS = SFCB2"
    puts $fd "PORT APUFCMLOADDVALID = FCB_LOADDVALID, DIR = IN, BUS = SFCB2" 
    puts $fd "PORT APUFCMLOADBYTEADDR = FCB_LOADBYTEADDR, DIR = IN, VEC = \[0:3\], BUS = SFCB2"
    puts $fd "PORT APUFCMENDIAN = FCB_ENDIAN, DIR = IN, BUS = SFCB2"
    puts $fd "PORT APUFCMDECUDI = FCB_DECUDI, DIR = IN, VEC = \[0:3\], BUS = SFCB2"
    puts $fd "PORT APUFCMDECUDIVALID = FCB_DECUDIVALID,  DIR = IN, BUS = SFCB2"
    puts $fd "PORT APUFCMDECFPUOP = FCB_DECFPUOP,  DIR = IN, BUS = SFCB2"
    puts $fd "PORT APUFCMDECLOAD = FCB_DECLOAD,  DIR = IN, BUS = SFCB2"
    puts $fd "PORT APUFCMDECSTORE = FCB_DECSTORE,  DIR = IN, BUS = SFCB2"
    puts $fd "PORT APUFCMDECLDSTXFERSIZE = FCB_DECLDSTXFERSIZE, DIR = IN, VEC = \[0:2\], BUS = SFCB2"
    puts $fd "PORT APUFCMDECNONAUTON = FCB_DECNONAUTON,  DIR = IN, BUS = SFCB2"
    puts $fd "PORT APUFCMNEXTINSTRREADY = FCB_NEXTINSTRREADY,  DIR = IN, BUS = SFCB2"
    puts $fd "PORT udidebug = \"\", DIR = OUT, VEC = \[153:0\]"
    puts $fd "PORT apu_clk = \"\", DIR = IN, SIGIS = CLK"
    puts $fd "PORT reset = FCB_RST, DIR = IN, BUS = SFCB2"
}

######################################################
# APU (PPC405) - SPECIFIC PROCS
######################################################

proc genApuGenerics {fd multibus hasSlave hasMaster} {
#    puts $fd "    C_PCMCIA_AWIDTH    : integer                   := 42;"
#    puts $fd "    C_PCMCIA_DWIDTH    : integer                   := 16;"
    
    if {! $multibus} {
#        puts $fd "    C_BASEADDR     : std_logic_vector(0 to C_PCMCIA_AWIDTH-1) := X\"3FFF0000000\""
    } else {
#        puts $fd "    C_PCMCIA_BASEADDR  : std_logic_vector(0 to C_PCMCIA_AWIDTH-1) := X\"3FFF0000000\""
    }
}

proc genApuPorts {fd hasSlave hasMaster} {
    puts $fd "    apufcminstruction : std_logic_vector(0 to 31);"
    puts $fd "    apufcminstrvalid : std_logic;"
    puts $fd "    apufcmradata : std_logic_vector(0 to 31);"
    puts $fd "    apufcmrbdata : std_logic_vector(0 to 31);"
    puts $fd "    apufcmoperandvalid : std_logic;"
    puts $fd "    apufcmflush : std_logic;"
    puts $fd "    apufcmwritebackok : std_logic;"
    puts $fd "    apufcmloaddata : std_logic_vector(0 to 31);"
    puts $fd "    apufcmloaddvalid : std_logic;"
    puts $fd "    apufcmloadbyteen : std_logic_vector(0 to 3);"
    puts $fd "    apufcmendian : std_logic;"
    puts $fd "    apufcmxerca : std_logic;"
    puts $fd "    apufcmdecoded : std_logic;"
    puts $fd "    apufcmdecudi : std_logic_vector(0 to 2);"
    puts $fd "    apufcmdecudivalid : std_logic;"
    puts $fd "    fcmapuinstrack : out std_logic;"
    puts $fd "    fcmapuresult : out std_logic_vector(0 to 31);"
    puts $fd "    fcmapudone : out std_logic;"
    puts $fd "    fcmapusleepnotready : out std_logic;"
    puts $fd "    fcmapudecodebusy : out std_logic;"
    puts $fd "    fcmapudcdgprwrite : out std_logic;"
    puts $fd "    fcmapudcdraen : out std_logic;"
    puts $fd "    fcmapudcdrben : out std_logic;"
    puts $fd "    fcmapudcdprivop : out std_logic;"
    puts $fd "    fcmapudcdforcealign : out std_logic;"
    puts $fd "    fcmapudcdxeroven : out std_logic;"
    puts $fd "    fcmapudcdxercaen : out std_logic;"
    puts $fd "    fcmapudcdcren : out std_logic;"
    puts $fd "    fcmapuexecrfield : out std_logic_vector(0 to 2);"
    puts $fd "    fcmapudcdload : out std_logic;"
    puts $fd "    fcmapudcdstore : out std_logic;"
    puts $fd "    fcmapudcdupdate : out std_logic;"
    puts $fd "    fcmapudcdldstbyte : out std_logic;"
    puts $fd "    fcmapudcdldsthw : out std_logic;"
    puts $fd "    fcmapudcdldstwd : out std_logic;"
    puts $fd "    fcmapudcdldstdw : out std_logic;"
    puts $fd "    fcmapudcdldstqw : out std_logic;"
    puts $fd "    fcmapudcdtraple : out std_logic;"
    puts $fd "    fcmapudcdtrapbe : out std_logic;"
    puts $fd "    fcmapudcdforcebesteering : out std_logic;"
    puts $fd "    fcmapudcdfpuop : out std_logic;"
    puts $fd "    fcmapuexeblockingmco : out std_logic;"
    puts $fd "    fcmapuexenonblockingmco : out std_logic;"
    puts $fd "    fcmapuloadwait : out std_logic;"
    puts $fd "    fcmapuresultvalid : out std_logic;"
    puts $fd "    fcmapuxerov : out std_logic;"
    puts $fd "    fcmapuxerca : out std_logic;"
    puts $fd "    fcmapucr : out std_logic_vector(0 to 3);"
    puts $fd "    fcmapuexception : out std_logic;"
    puts -nonewline $fd "    apu_clk, reset : in std_logic"
}

proc genApuConnSignals {fd con} {
    set type [lindex $con 0]
    if {$type != "memory" && $type != "instruction"} {
	set iname [lindex $con 2]
        puts $fd "  signal ${iname}_result : std_logic_vector(0 to 31);"
        puts $fd "  signal ${iname}_done : std_logic;"
        puts $fd "  signal ${iname}_sleepnotready : std_logic;"
        puts $fd "  signal ${iname}_loadwait : std_logic;"
        puts $fd "  signal ${iname}_resultvalid : std_logic;"
        puts $fd "  signal ${iname}_cs : std_logic;"
        puts $fd "  signal ${iname}_instrvalid : std_logic;"
    }
}

proc genApuSignals {fd noAttachments} {
    set selWidth [calcAddrBits $noAttachments]
    set addrWidth [expr $selWidth + 4]

    if {$selWidth > 0} {
	puts $fd "  signal apusel : std_logic_vector(0 to [expr $selWidth - 1]);"

    }
    puts $fd "  signal apufcmdecfpuop : std_logic;"
}

proc genApuIpif {fd multibus noAttachments} {
    set selWidth [calcAddrBits $noAttachments]
    set firstSelBit [expr 9 - $selWidth]

    # bits 6-10 of the instruction are the "RT" field, and bits 9,10 are internal to a stream.
    # this leaves stream-selection to bits 6-8 and thus limits us to 8 streams.
    puts $fd "apusel <= apufcminstruction($firstSelBit to 8);"

    # ignore FPU load/stores
    puts $fd "  apufcmdecfpuop <= '1' when apufcminstruction(0)='1' or apufcminstruction(26)='1' else '0';"
}

# generates code to OR together the output signals from each stream.
# also ties all the unused fcm->apu lines to ground.
proc genApuOutputSignals {fd cons} {
#    puts $fd "cons has $[llength $cons] ([lindex $cons 0])"
	
    foreach output [list result done sleepnotready loadwait resultvalid] {
        # is there a simple, clean way to avoid or-ing with zero?
        puts -nonewline $fd "  fcmapu${output} <= "
        set iter 0
        foreach con $cons {
            set type [lindex $con 0]
            set dir [lindex $con 1]
            set iname [lindex $con 2]
            set width [lindex $con 3]
				set busName [lindex $con 4]
				if {[selectBus $con] == $busName} {
					if {$iter > 0} {
						puts -nonewline $fd " or ";
					}
					puts -nonewline $fd "${iname}_${output}";
					incr iter
				}
        }
        puts $fd ";"
    }
    puts $fd "\n  -- unused signals"
    puts $fd "  FCMAPUINSTRACK <= '0';"
    puts $fd "  FCMAPUDECODEBUSY <= '0';"
    puts $fd "  FCMAPUDCDGPRWRITE <= '0';"
    puts $fd "  FCMAPUDCDRAEN <= '0';"
    puts $fd "  FCMAPUDCDRBEN <= '0';"
    puts $fd "  FCMAPUDCDPRIVOP <= '0';"
    puts $fd "  FCMAPUDCDFORCEALIGN <= '0';"
    puts $fd "  FCMAPUDCDXEROVEN <= '0';"
    puts $fd "  FCMAPUDCDXERCAEN <= '0';"
    puts $fd "  FCMAPUDCDCREN <= '0';"
    puts $fd "  FCMAPUEXECRFIELD <= \"000\";"
    puts $fd "  FCMAPUDCDLOAD <= '0';"
    puts $fd "  FCMAPUDCDSTORE <= '0';"
    puts $fd "  FCMAPUDCDUPDATE <= '0';"
    puts $fd "  FCMAPUDCDLDSTBYTE <= '0';"
    puts $fd "  FCMAPUDCDLDSTHW <= '0';"
    puts $fd "  FCMAPUDCDLDSTWD <= '0';"
    puts $fd "  FCMAPUDCDLDSTDW <= '0';"
    puts $fd "  FCMAPUDCDLDSTQW <= '0';"
    puts $fd "  FCMAPUDCDTRAPLE <= '0';"
    puts $fd "  FCMAPUDCDTRAPBE <= '0';"
    puts $fd "  FCMAPUDCDFORCEBESTEERING <= '0';"
    puts $fd "  FCMAPUDCDFPUOP <= '0';"
    puts $fd "  FCMAPUEXEBLOCKINGMCO <= '0';"
    puts $fd "  FCMAPUEXENONBLOCKINGMCO <= '0';"
    puts $fd "  FCMAPUXEROV <= '0';"
    puts $fd "  FCMAPUXERCA <= '0';"
    puts $fd "  FCMAPUCR <= \"0000\";"
    puts $fd "  FCMAPUEXCEPTION <= '0';"
    puts $fd ""
}

proc genApuConn {fd con no total rstName clkName} {
    set selWidth [calcAddrBits $total]
    
    set type [lindex $con 0]
    set dir [lindex $con 1]
    set iname [lindex $con 2]
    set width [lindex $con 3]
    switch $type {
        {stream} {
	    set addr [makeVec $no $selWidth]
 	        puts $fd "  ${iname}_cs <= '1' when apusel = $addr else '0';"
            puts $fd "  ${iname}_instrvalid <= '1' when (apufcminstrvalid = '1') and (${iname}_cs = '1') and apufcmdecfpuop = '0' else '0';"
	    if {$dir == "in"} {
	        puts $fd "  ${iname}_if: apu_to_stream"
	    } else {
	        puts $fd "  ${iname}_if: stream_to_apu"
	    }
	    puts $fd "    generic map ("
	    puts $fd "      datawidth => $width"
	    puts $fd "    )"
            puts $fd "    port map ("
            puts $fd "      reset, apu_clk,"
            puts $fd "      ${iname}_cs,"
            puts $fd "      apufcminstruction,"
            puts $fd "      ${iname}_instrvalid,"
            puts $fd "      apufcmradata,"
            puts $fd "      apufcmrbdata,"
            puts $fd "      apufcmoperandvalid,"
            puts $fd "      apufcmflush,"
            puts $fd "      apufcmwritebackok,"
            puts $fd "      apufcmloaddata,"
            puts $fd "      apufcmloaddvalid,"
            puts $fd "      apufcmloadbyteen,"
            puts $fd "      apufcmendian,"
            puts $fd "      apufcmxerca,"
            puts $fd "      apufcmdecoded,"
            puts $fd "      apufcmdecudi,"
            puts $fd "      apufcmdecudivalid,"
            puts $fd "      ${iname}_result,"
            puts $fd "      ${iname}_done,"
            puts $fd "      ${iname}_sleepnotready,"
            puts $fd "      ${iname}_loadwait,"
            puts $fd "      ${iname}_resultvalid,"
            puts $fd "      ${iname}_en,"
            puts $fd "      ${iname}_rdy,"
            puts $fd "      ${iname}_eos,"
            puts $fd "      ${iname}_idata"
            puts $fd "    );\n"
        }
        {signal} {
            error "TODO"
        }
        {register} {
            error "TODO"
	}
        {memory} {
	    error "TODO"
        }
    }
}

proc genApuMPD {fd} {
    puts $fd "OPTION IMP_NETLIST = TRUE"
    puts $fd "OPTION SIM_MODELS = BEHAVIORAL"
    puts $fd "OPTION IP_GROUP = MICROBLAZE:PPC:USER"

    puts $fd "OPTION RUN_NGCBUILD = TRUE"

    puts $fd "BUS_INTERFACE BUS = SFCB, BUS_STD = FCB, BUS_TYPE = SLAVE"

    puts $fd "PORT FCMAPUINSTRACK = Sl_INSTRACK, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPURESULT = Sl_RESULT, DIR = OUT, VEC = \[0:31\], BUS = SFCB"
    puts $fd "PORT FCMAPUDONE = Sl_DONE,DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUSLEEPNOTREADY = Sl_SLEEPNOTREADY, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDECODEBUSY = Sl_DECODEBUSY, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDGPRWRITE = Sl_DCDGPRWRITE, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDRAEN = Sl_DCDRAEN, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDRBEN = Sl_DCDRBEN, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDPRIVOP = Sl_DCDPRIVOP, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDFORCEALIGN = Sl_DCDFORCEALIGN,DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDXEROVEN = Sl_DCDXEROVEN, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDXERCAEN = Sl_DCDXERCAEN, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDCREN = Sl_DCDCREN,DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUEXECRFIELD = Sl_EXECRFIELD, DIR = OUT, VEC = \[0:2\], BUS = SFCB"
    puts $fd "PORT FCMAPUDCDLOAD = Sl_DCDLOAD, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDSTORE = Sl_DCDSTORE, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDUPDATE = Sl_DCDUPDATE, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDLDSTBYTE = Sl_DCDLDSTBYTE, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDLDSTHW = Sl_DCDLDSTHW, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDLDSTWD = Sl_DCDLDSTWD, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDLDSTDW = Sl_DCDLDSTDW, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDLDSTQW = Sl_DCDLDSTQW, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDTRAPLE = Sl_DCDTRAPLE, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDTRAPBE = Sl_DCDTRAPBE, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDFORCEBESTEERING = Sl_DCDFORCEBESTEERING, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUDCDFPUOP = Sl_DCDFPUOP, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUEXEBLOCKINGMCO = Sl_EXEBLOCKINGMCO, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUEXENONBLOCKINGMCO = Sl_EXENONBLOCKINGMCO, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPULOADWAIT = Sl_LOADWAIT, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPURESULTVALID = Sl_RESULTVALID, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUXEROV = Sl_XEROV, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUXERCA = Sl_XERCA, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT FCMAPUCR = Sl_CR, DIR = OUT, VEC = \[0:3\], BUS = SFCB"
    puts $fd "PORT FCMAPUEXCEPTION = Sl_EXCEPTION, DIR = OUT, BUS = SFCB" 
    puts $fd "PORT APUFCMINSTRUCTION = FCB_INSTRUCTION, DIR = IN, VEC = \[0:31\], BUS = SFCB"
    puts $fd "PORT APUFCMINSTRVALID = FCB_INSTRVALID, DIR = IN, BUS = SFCB" 
    puts $fd "PORT APUFCMRADATA = FCB_RADATA, DIR = IN, VEC = \[0:31\], BUS = SFCB"
    puts $fd "PORT APUFCMRBDATA = FCB_RBDATA, DIR = IN, VEC = \[0:31\], BUS = SFCB"
    puts $fd "PORT APUFCMOPERANDVALID = FCB_OPERANDVALID,  DIR = IN, BUS = SFCB" 
    puts $fd "PORT APUFCMFLUSH = FCB_FLUSH, DIR = IN, BUS = SFCB" 
    puts $fd "PORT APUFCMWRITEBACKOK = FCB_WRITEBACKOK, DIR = IN, BUS = SFCB" 
    puts $fd "PORT APUFCMLOADDATA = FCB_LOADDATA, DIR = IN, VEC = \[0:31\], BUS = SFCB"
    puts $fd "PORT APUFCMLOADDVALID = FCB_LOADDVALID, DIR = IN, BUS = SFCB" 
    puts $fd "PORT APUFCMLOADBYTEEN = FCB_LOADBYTEEN, DIR = IN, VEC = \[0:3\], BUS = SFCB"
    puts $fd "PORT APUFCMENDIAN = FCB_ENDIAN, DIR = IN, BUS = SFCB"
    puts $fd "PORT APUFCMXERCA = FCB_XERCA, DIR = IN, BUS = SFCB" 
    puts $fd "PORT APUFCMDECODED = FCB_DECODED, DIR = IN, BUS = SFCB"
    puts $fd "PORT APUFCMDECUDI = FCB_DECUDI, DIR = IN, VEC = \[0:2\], BUS = SFCB"
    puts $fd "PORT APUFCMDECUDIVALID = FCB_DECUDIVALID,  DIR = IN, BUS = SFCB"
    puts $fd "PORT apu_clk = \"\", DIR = IN, SIGIS = CLK"
    puts $fd "PORT reset = FCB_RST, DIR = IN, BUS = SFCB"
}

######################################################
# PCMCIA-SPECIFIC PROCS
######################################################

proc genPcmciaGenerics {fd multibus hasSlave hasMaster} {
    puts $fd "    C_PCMCIA_AWIDTH    : integer                   := 42;"
    puts $fd "    C_PCMCIA_DWIDTH    : integer                   := 16;"
    
    if {! $multibus} {
        puts $fd "    C_BASEADDR     : std_logic_vector(0 to C_PCMCIA_AWIDTH-1) := X\"3FFF0000000\""
    } else {
        puts $fd "    C_PCMCIA_BASEADDR  : std_logic_vector(0 to C_PCMCIA_AWIDTH-1) := X\"3FFF0000000\""
    }
}

proc genPcmciaPorts {fd hasSlave hasMaster} {
    puts -nonewline $fd    "    PCMCIA_Clk       : in std_logic"
    puts -nonewline $fd ";\n    PCMCIA_Addr      : in std_logic_vector(0 to C_PCMCIA_AWIDTH-1)"
    puts -nonewline $fd ";\n    PCMCIA_wrDBus    : in std_logic_vector(0 to C_PCMCIA_DWIDTH-1)"
    puts -nonewline $fd ";\n    Sl_rdDBus        : out std_logic_vector(0 to C_PCMCIA_DWIDTH-1)"
    puts -nonewline $fd ";\n    PCMCIA_Rd        : in std_logic"
    puts -nonewline $fd ";\n    PCMCIA_Wr        : in std_logic"
}

proc genPcmciaConnSignals {fd con} {
    set type [lindex $con 0]
    set iname [lindex $con 2]
    if {$type != "memory"} {
        puts $fd "  signal ${iname}_cs : std_logic;"
        puts $fd "  signal ${iname}_odata : std_logic_vector(15 downto 0);"
    }
}

proc genPcmciaSignals {fd noAttachments} {
    set selWidth [calcAddrBits $noAttachments]
    set addrWidth [expr $selWidth + 4]

    puts $fd "  signal IP2PCMCIA_Data : std_logic_vector (C_PCMCIA_DWIDTH-1 downto 0);"
    #puts $fd "  signal PCMCIA2IP_Clk : std_logic;"
    puts $fd "  signal PCMCIA2IP_CS   : std_logic;"
    puts $fd "  signal PCMCIA2IP_Rd : std_logic;"
    puts $fd "  signal PCMCIA2IP_Wr : std_logic;"
    puts $fd "  signal PCMCIA2IP_Addr : std_logic_vector ([expr $addrWidth - 1] downto 0);"
    puts $fd "  signal PCMCIA2IP_Data : std_logic_vector (C_PCMCIA_DWIDTH-1 downto 0);"
    puts $fd "  signal PCMCIA2IP_Reset : std_logic;"
    if {$selWidth > 0} {
	puts $fd "  signal pcmciasel : std_logic_vector([expr $selWidth - 1] downto 0);"
    }

    # component decls for stuff in pcmcia_if.v
    puts $fd "\n  component pico_pcmcia_ipif is"
    puts $fd "  generic (C_BASEADDR   : std_logic_vector(0 to C_PCMCIA_AWIDTH-1);"
    puts $fd "           C_PCMCIA_AWIDTH : positive;"
    puts $fd "           C_IMPULSE_AWIDTH: positive);"
    puts $fd "  port (clk : in std_ulogic;"
    puts $fd "        pcmcia_addr     : in std_logic_vector(C_PCMCIA_AWIDTH-1 downto 0);"
    puts $fd "        pcmcia_read     : in std_logic;"
    puts $fd "        pcmcia_write    : in std_logic;"
    puts $fd "        pcmcia_din      : in std_logic_vector(C_PCMCIA_DWIDTH-1 downto 0);"
    puts $fd "        pcmcia_dout     : out std_logic_vector(C_PCMCIA_DWIDTH-1 downto 0);"
    puts $fd "        impulse_addr    : out std_logic_vector([expr $addrWidth - 1] downto 0);"
    puts $fd "        impulse_read    : out std_logic;"
    puts $fd "        impulse_write   : out std_logic;"
    puts $fd "        impulse_din     : out std_logic_vector(C_PCMCIA_DWIDTH-1 downto 0);"
    puts $fd "        impulse_dout    : in  std_logic_vector(C_PCMCIA_DWIDTH-1 downto 0));"
    puts $fd "  end component pico_pcmcia_ipif;\n"
    puts $fd "  component stream_to_pcmcia is"
    puts $fd "  generic (datawidth : positive := 8);"
    puts $fd "  port (clk : in std_logic;"
    puts $fd "        pcmcia_addr : in std_logic_vector(3 downto 0);"
    puts $fd "        pcmcia_en : in std_logic;"
    puts $fd "        pcmcia_rd : in std_logic;"
    puts $fd "        pcmcia_wr : in std_logic;"
    puts $fd "        pcmcia_wdata : in std_logic_vector(C_PCMCIA_DWIDTH-1 downto 0);"
    puts $fd "        pcmcia_rdata : out std_logic_vector(C_PCMCIA_DWIDTH-1 downto 0);"
    puts $fd "        stream_en : out std_ulogic;"
    puts $fd "        stream_rdy : in std_ulogic;"
    puts $fd "        stream_eos : in std_ulogic;"
    puts $fd "        stream_data : in std_ulogic_vector(datawidth-1 downto 0));"
    puts $fd "  end component stream_to_pcmcia;\n"
    puts $fd "  component pcmcia_to_stream is"
    puts $fd "  generic (datawidth : positive := 8);"
    puts $fd "  port (clk : in std_logic;"
    puts $fd "        pcmcia_addr : in std_logic_vector(3 downto 0);"
    puts $fd "        pcmcia_en : in std_logic;"
    puts $fd "        pcmcia_rd : in std_logic;"
    puts $fd "        pcmcia_wr : in std_logic;"
    puts $fd "        pcmcia_wdata : in std_logic_vector(C_PCMCIA_DWIDTH-1 downto 0);"
    puts $fd "        pcmcia_rdata : out std_logic_vector(C_PCMCIA_DWIDTH-1 downto 0);"
    puts $fd "        stream_en : out std_ulogic;"
    puts $fd "        stream_rdy : in std_ulogic;"
    puts $fd "        stream_eos : out std_ulogic;"
    puts $fd "        stream_data : out std_ulogic_vector(datawidth-1 downto 0));"
    puts $fd "  end component pcmcia_to_stream;\n"
}

proc genPcmciaIpif {fd multibus noAttachments} {
    set selWidth [calcAddrBits $noAttachments]
    set addrWidth [expr $selWidth + 4]

    puts $fd "  ipif : pico_pcmcia_ipif"
    puts $fd "  generic map"
    puts $fd "  ("
    if {$multibus} {
        puts $fd "    C_BASEADDR => C_PCMCIA_BASEADDR,"
    } else {
        puts $fd "    C_BASEADDR => C_BASEADDR,"
    }
    puts $fd "    C_PCMCIA_AWIDTH => C_PCMCIA_AWIDTH,"
    puts $fd "    C_IMPULSE_AWIDTH => $addrWidth"
    puts $fd "  )"
    puts $fd "  port map"
    puts $fd "  ("
    puts $fd "    PCMCIA_Clk,"
    puts $fd "    PCMCIA_Addr,"
    puts $fd "    PCMCIA_Rd,"
    puts $fd "    PCMCIA_Wr,"
    puts $fd "    PCMCIA_wrDBus,"
    puts $fd "    Sl_rdDBus,"
    puts $fd "    PCMCIA2IP_Addr,"
    puts $fd "    PCMCIA2IP_Rd,"
    puts $fd "    PCMCIA2IP_Wr,"
    puts $fd "    PCMCIA2IP_Data,"
    puts $fd "    IP2PCMCIA_Data"
    puts $fd "  );"
    puts $fd ""
}

proc genPcmciaConn {fd con no total rstName clkName} {
    set selWidth [calcAddrBits $total]
    set addrWidth [expr $selWidth + 4]
    set streamsel "([expr $addrWidth - 1] downto 4)"
    
    set type [lindex $con 0]
    set dir [lindex $con 1]
    set iname [lindex $con 2]
    set width [lindex $con 3]
    switch $type {
        {stream} {
	    set addr [makeVec $no $selWidth]
            if {$selWidth} {
                puts $fd "  ${iname}_cs <="
                puts $fd "    '1' when PCMCIA2IP_Addr${streamsel} = $addr and PCMCIA2IP_Rd = '1' else"
                puts $fd "    '1' when PCMCIA2IP_Addr${streamsel} = $addr and PCMCIA2IP_Wr = '1' else"
                puts $fd "    '0';\n"
            } else {
                puts $fd "  ${iname}_cs <= PCMCIA2IP_Rd or PCMCIA2IP_Wr;\n"
            }
	    if {$dir == "in"} {
	        puts $fd "  ${iname}_if: pcmcia_to_stream"
	    } else {
	        puts $fd "  ${iname}_if: stream_to_pcmcia"
	    }
	    puts $fd "    generic map ("
	    puts $fd "      datawidth => $width"
	    puts $fd "    )"
	    puts $fd "    port map ("
	    puts $fd "      clk => PCMCIA_Clk,"
	    puts $fd "      pcmcia_addr => PCMCIA2IP_addr(3 downto 0),"
	    puts $fd "      pcmcia_en => ${iname}_cs,"
	    puts $fd "      pcmcia_rd => PCMCIA2IP_Rd,"
	    puts $fd "      pcmcia_wr => PCMCIA2IP_Wr,"
	    puts $fd "      pcmcia_wdata => PCMCIA2IP_Data,"
	    puts $fd "      pcmcia_rdata => ${iname}_odata,"
	    puts $fd "      stream_en => ${iname}_en,"
	    puts $fd "      stream_rdy => ${iname}_rdy,"
	    puts $fd "      stream_eos => ${iname}_eos,"
	    puts $fd "      stream_data => ${iname}_idata"
	    puts $fd "    );\n"
        }
        {signal} {
            error "TODO"
	    set addr [makeVec $no $selWidth]
 	    puts $fd "  ${iname}_cs <="
	    puts $fd "    '1' when Plb2IP_raddr${streamsel} = $addr and Plb2IP_RdCE = '1' else"
	    puts $fd "    '1' when Plb2IP_waddr${streamsel} = $addr and Plb2IP_WrCE = '1' else"
	    puts $fd "    '0';\n"
	    if {$dir == "in"} {
	        puts $fd "  ${iname}_if: plb_to_signal"
	    } else {
	        puts $fd "  ${iname}_if: signal_to_plb"
	    }
	    puts $fd "    generic map ("
	    puts $fd "      datawidth => $width"
	    puts $fd "    )"
	    puts $fd "    port map ("
	    puts $fd "      clk => Plb2IP_Clk,"
	    puts $fd "      plb_raddr => Plb2IP_raddr(3 downto 3),"
	    puts $fd "      plb_waddr => Plb2IP_waddr(3 downto 3),"
	    puts $fd "      plb_ce => ${iname}_cs,"
	    puts $fd "      plb_read => Plb2IP_RdCe,"
	    puts $fd "      plb_reset => Plb2IP_Reset,"
	    puts $fd "      plb_write => Plb2IP_WrCe,"
	    puts $fd "      plb_wdata => Plb2IP_Data,"
	    puts $fd "      plb_rdata => ${iname}_odata,"
	    puts $fd "      signal_en => ${iname}_en,"
	    if {$dir == "out"} {
	      puts $fd "      signal_rdy => ${iname}_rdy,"
            }
	    puts $fd "      signal_data => ${iname}_data"
	    puts $fd "    );\n"
        }
        {register} {
            error "TODO"
	    set addr [makeVec $no $selWidth]
 	    puts $fd "  ${iname}_cs <="
	    puts $fd "    '1' when Plb2IP_raddr${streamsel} = $addr and Plb2IP_RdCE = '1' else"
	    puts $fd "    '1' when Plb2IP_waddr${streamsel} = $addr and Plb2IP_WrCE = '1' else"
	    puts $fd "    '0';\n"
	    if {$dir == "in"} {
	        puts $fd "  ${iname}_if: plb_to_register"
	    } else {
	        puts $fd "  ${iname}_if: register_to_plb"
	    }
	    puts $fd "    generic map ("
	    puts $fd "      datawidth => $width"
	    puts $fd "    )"
	    puts $fd "    port map ("
	    puts $fd "      clk => Plb2IP_Clk,"
	    puts $fd "      plb_raddr => Plb2IP_raddr(3 downto 3),"
	    puts $fd "      plb_waddr => Plb2IP_waddr(3 downto 3),"
	    puts $fd "      plb_ce => ${iname}_cs,"
	    puts $fd "      plb_read => Plb2IP_RdCe,"
	    puts $fd "      plb_reset => Plb2IP_Reset,"
	    puts $fd "      plb_write => Plb2IP_WrCe,"
	    puts $fd "      plb_wdata => Plb2IP_Data,"
	    puts $fd "      plb_rdata => ${iname}_odata,"
	    if {$dir == "in"} {
	      puts $fd "      register_en => ${iname}_en,"
	      puts $fd "      register_data => ${iname}_data"
            } else {
	      puts $fd "      register_value => ${iname}_value"
            }
	    puts $fd "    );\n"
	}
        {memory} {
	    error "TODO"
        }
    }
}

######################################################
# PLB SPECIFIC PROCS
######################################################

proc genPlbGenerics {fd multibus hasSlave hasMaster v4} {
    if {! $multibus} {
        puts $fd "    C_BASEADDR     : std_logic_vector(0 to 31) := X\"FFFFFFFF\";"
        puts $fd "    C_HIGHADDR     : std_logic_vector(0 to 31) := X\"00000000\";"
    } else {
        puts $fd "    C_PLB_BASEADDR     : std_logic_vector(0 to 31) := X\"FFFFFFFF\";"
        puts $fd "    C_PLB_HIGHADDR     : std_logic_vector(0 to 31) := X\"00000000\";"
    }
    if {$v4} {
	puts $fd "    C_SPLB_NUM_MASTERS  : integer                   := 8;"
	puts $fd "    C_SPLB_MID_WIDTH    : integer                   := 3;"
	puts $fd "    C_SPLB_AWIDTH       : integer                   := 32;"
	puts -nonewline $fd "    C_SPLB_DWIDTH       : integer                   := 64"
	if {$hasMaster} {
	    puts $fd ";\n    C_MPLB_AWIDTH       : integer                   := 32;"
	    puts -nonewline $fd "    C_MPLB_DWIDTH       : integer                   := 64"
	}
    } else {
    puts $fd "    C_PLB_NUM_MASTERS  : integer                   := 8;"
    puts $fd "    C_PLB_MID_WIDTH    : integer                   := 3;"
    puts $fd "    C_PLB_AWIDTH       : integer                   := 32;"
    puts -nonewline $fd "    C_PLB_DWIDTH       : integer                   := 64"
}
}

proc genPlbPorts {fd hasSlave hasMaster v4} {
    if {$v4} {
	set slaveGeneric "C_SPLB"
	set masterGeneric "C_MPLB"
    } else {
	set slaveGeneric "C_PLB"
	set masterGeneric "C_PLB"
    }
    puts -nonewline $fd "    PLB_Clk          : in std_logic"
    puts -nonewline $fd ";\n    PLB_Rst          : in std_logic"
    puts -nonewline $fd ";\n    PLB_abort        : in std_logic"
    puts -nonewline $fd ";\n    PLB_ABus         : in std_logic_vector(0 to ${slaveGeneric}_AWIDTH-1)"
    puts -nonewline $fd ";\n    PLB_BE           : in std_logic_vector(0 to (${slaveGeneric}_DWIDTH/8)-1)"
    puts -nonewline $fd ";\n    PLB_busLock      : in std_logic"
    if {$v4} {
	puts -nonewline $fd ";\n    PLB_TAttribute   : in std_logic_vector(0 to 15)"
    } else {
    puts -nonewline $fd ";\n    PLB_compress     : in std_logic"
    puts -nonewline $fd ";\n    PLB_guarded      : in std_logic"
    puts -nonewline $fd ";\n    PLB_ordered      : in std_logic"
    }
    puts -nonewline $fd ";\n    PLB_lockErr      : in std_logic"
    puts -nonewline $fd ";\n    PLB_masterID     : in  std_logic_vector(0 to ${slaveGeneric}_MID_WIDTH-1)"
    puts -nonewline $fd ";\n    PLB_MSize        : in std_logic_vector(0 to 1)"
    puts -nonewline $fd ";\n    PLB_PAValid      : in std_logic"
    puts -nonewline $fd ";\n    PLB_RNW          : in std_logic"
    puts -nonewline $fd ";\n    PLB_size         : in std_logic_vector(0 to 3)"
    puts -nonewline $fd ";\n    PLB_type         : in std_logic_vector(0 to 2)"
    if {$v4} {
	puts -nonewline $fd ";\n    PLB_UABus        : in std_logic_vector(0 to ${slaveGeneric}_AWIDTH-1)"
    }
    puts -nonewline $fd ";\n    Sl_addrAck       : out std_logic"
    puts -nonewline $fd ";\n    Sl_MBusy         : out std_logic_vector(0 to ${slaveGeneric}_NUM_MASTERS-1)"
    if {$v4} {
	puts -nonewline $fd ";\n    Sl_MRdErr        : out std_logic_vector(0 to ${slaveGeneric}_NUM_MASTERS-1)"
	puts -nonewline $fd ";\n    Sl_MWrErr        : out std_logic_vector(0 to ${slaveGeneric}_NUM_MASTERS-1)"
    } else {
	puts -nonewline $fd ";\n    Sl_MErr          : out std_logic_vector(0 to ${slaveGeneric}_NUM_MASTERS-1)"
    }
    puts -nonewline $fd ";\n    Sl_rearbitrate   : out std_logic"
    puts -nonewline $fd ";\n    Sl_SSize         : out std_logic_vector(0 to 1)"
    puts -nonewline $fd ";\n    Sl_wait          : out std_logic"
    if {$v4} {
	puts -nonewline $fd ";\n    Sl_MIRQ          : out std_logic_vector(0 to ${slaveGeneric}_NUM_MASTERS-1)"
    }
    puts -nonewline $fd ";\n    PLB_rdPrim       : in std_logic"
    puts -nonewline $fd ";\n    PLB_SAValid      : in std_logic"
    puts -nonewline $fd ";\n    PLB_wrPrim       : in std_logic"
    puts -nonewline $fd ";\n    PLB_wrBurst      : in std_logic"
    puts -nonewline $fd ";\n    PLB_wrDBus       : in std_logic_vector(0 to ${slaveGeneric}_DWIDTH-1)"
    puts -nonewline $fd ";\n    Sl_wrBTerm       : out std_logic"
    puts -nonewline $fd ";\n    Sl_wrComp        : out std_logic"
    puts -nonewline $fd ";\n    Sl_wrDAck        : out std_logic"
    puts -nonewline $fd ";\n    PLB_rdBurst      : in std_logic"
    puts -nonewline $fd ";\n    Sl_rdBTerm       : out std_logic"
    puts -nonewline $fd ";\n    Sl_rdComp        : out std_logic"
    puts -nonewline $fd ";\n    Sl_rdDAck        : out std_logic"
    puts -nonewline $fd ";\n    Sl_rdDBus        : out std_logic_vector(0 to ${slaveGeneric}_DWIDTH-1)"
    puts -nonewline $fd ";\n    Sl_rdWdAddr      : out std_logic_vector(0 to 3)"
    if {$v4} {
	puts -nonewline $fd ";\n    PLB_rdPendReq    : in  std_logic"
	puts -nonewline $fd ";\n    PLB_wrPendReq    : in  std_logic"
	puts -nonewline $fd ";\n    PLB_rdPendPri    : in  std_logic_vector(0 to 1)"
	puts -nonewline $fd ";\n    PLB_wrPendPri    : in  std_logic_vector(0 to 1)"
    } else {
    puts -nonewline $fd ";\n    PLB_pendReq      : in  std_logic"
    puts -nonewline $fd ";\n    PLB_pendPri      : in  std_logic_vector(0 to 1)"
    }
    puts -nonewline $fd ";\n    PLB_reqPri       : in  std_logic_vector(0 to 1)"
    if {$hasMaster} {
	puts -nonewline $fd ";\n    PLB_MAddrAck            : in  std_logic"
	puts -nonewline $fd ";\n    PLB_MSSize              : in  std_logic_vector(0 to 1)"
	puts -nonewline $fd ";\n    PLB_MRearbitrate        : in  std_logic"
	puts -nonewline $fd ";\n    PLB_MBusy               : in  std_logic"
	if {$v4} {
	    puts -nonewline $fd ";\n    PLB_MRdErr                : in  std_logic"
	    puts -nonewline $fd ";\n    PLB_MWrErr                : in  std_logic"	} else {
	puts -nonewline $fd ";\n    PLB_MErr                : in  std_logic"
	}
	puts -nonewline $fd ";\n    PLB_MWrDAck             : in  std_logic"
	puts -nonewline $fd ";\n    PLB_MRdDBus             : in  std_logic_vector(0 to (${masterGeneric}_DWIDTH-1))"
	puts -nonewline $fd ";\n    PLB_MRdWdAddr           : in  std_logic_vector(0 to 3)"
	puts -nonewline $fd ";\n    PLB_MRdDAck             : in  std_logic"
	puts -nonewline $fd ";\n    PLB_MRdBTerm            : in  std_logic"
	puts -nonewline $fd ";\n    PLB_MWrBTerm            : in  std_logic"
	if {$v4} {
	    puts -nonewline $fd ";\n    PLB_MTimeout            : in  std_logic"
	    puts -nonewline $fd ";\n    PLB_MIRQ                : in  std_logic"
	}
	puts -nonewline $fd ";\n    M_abort      : out  std_logic"
	puts -nonewline $fd ";\n    M_ABus         : out std_logic_vector(0 to 31)"
	puts -nonewline $fd ";\n    M_BE           : out std_logic_vector(0 to (${masterGeneric}_DWIDTH/8)-1)"
	puts -nonewline $fd ";\n    M_busLock      : out  std_logic"
	if {$v4} {
	    puts -nonewline $fd ";\n    M_TAttribute   : out  std_logic_vector(0 to 15)"
	} else {
	puts -nonewline $fd ";\n    M_compress      : out  std_logic"
	puts -nonewline $fd ";\n    M_guarded      : out  std_logic"
	}
	puts -nonewline $fd ";\n    M_lockErr      : out  std_logic"
	puts -nonewline $fd ";\n    M_MSize           : out std_logic_vector(0 to 1)"
	if {!$v4} {
	puts -nonewline $fd ";\n    M_ordered      : out  std_logic"
	}
	puts -nonewline $fd ";\n    M_priority           : out std_logic_vector(0 to 1)"
	puts -nonewline $fd ";\n    M_rdBurst      : out  std_logic"
	puts -nonewline $fd ";\n    M_request      : out  std_logic"
	puts -nonewline $fd ";\n    M_RNW      : out  std_logic"
	puts -nonewline $fd ";\n    M_size           : out std_logic_vector(0 to 3)"
	puts -nonewline $fd ";\n    M_type           : out std_logic_vector(0 to 2)"
	puts -nonewline $fd ";\n    M_wrBurst      : out  std_logic"
	puts -nonewline $fd ";\n    M_wrDBus        : out std_logic_vector(0 to ${masterGeneric}_DWIDTH-1)"
	if {$v4} {
	    puts -nonewline $fd ";\n    M_UABus         : out std_logic_vector(0 to 31)"
	}
    }
}

proc genPlbConnSignals {fd con} {
    set type [lindex $con 0]
    set iname [lindex $con 2]
    if {$type != "memory"} {
        puts $fd "  signal ${iname}_cs : std_logic;"
        puts $fd "  signal ${iname}_odata : std_logic_vector(C_SPLB_DWIDTH-1 downto 0);"
    }
}

proc genPlbSignals {fd noAttachments v4} {
    set selWidth [calcAddrBits $noAttachments]
    set addrWidth [expr $selWidth + 4]

    if {$v4} {
	puts $fd "  signal IP2Plb_data : std_logic_vector (C_SPLB_DWIDTH-1 downto 0);"
    } else {
    puts $fd "  signal IP2Plb_data : std_logic_vector (C_PLB_DWIDTH-1 downto 0);"
    }
    puts $fd "  signal Plb2IP_Reset : std_logic;"
    puts $fd "  signal Plb2IP_Clk : std_logic;"
    puts $fd "  signal Plb2IP_CS : std_logic;"
    puts $fd "  signal Plb2IP_RdCE : std_logic;"
    puts $fd "  signal Plb2IP_WrCE : std_logic;"
    puts $fd "  signal Plb2IP_raddr : std_logic_vector ([expr $addrWidth - 1] downto 0);"
    puts $fd "  signal Plb2IP_waddr : std_logic_vector ([expr $addrWidth - 1] downto 0);"
    if {$v4} {
	puts $fd "  signal Plb2IP_data : std_logic_vector (C_SPLB_DWIDTH-1 downto 0);"
    } else {
    puts $fd "  signal Plb2IP_data : std_logic_vector (C_PLB_DWIDTH-1 downto 0);"
    }
    if {$selWidth > 0} {
	puts $fd "  signal plbsel : std_logic_vector([expr $selWidth - 1] downto 0);"
    }
}

proc genPlbIpif {fd multibus noAttachments v4} {
    set selWidth [calcAddrBits $noAttachments]
    set addrWidth [expr $selWidth + 4]

    puts $fd "  ipif : impulse_plb_ipif"
    puts $fd "  generic map"
    puts $fd "  ("
    if {$multibus} {
        puts $fd "    C_PLB_BASEADDR, C_PLB_HIGHADDR,"
    } else {
        puts $fd "    C_BASEADDR, C_HIGHADDR,"
    }
    if {$v4} {
	puts $fd "    C_SPLB_NUM_MASTERS,"
	puts $fd "    C_SPLB_MID_WIDTH,"
	puts $fd "    C_SPLB_AWIDTH,"
	puts $fd "    C_SPLB_DWIDTH,"
    } else {
    puts $fd "    C_PLB_NUM_MASTERS,"
    puts $fd "    C_PLB_MID_WIDTH,"
    puts $fd "    C_PLB_AWIDTH,"
    puts $fd "    C_PLB_DWIDTH,"
    }
    puts $fd "    $addrWidth"
    puts $fd "  )"
    puts $fd "  port map"
    puts $fd "  ("
    puts $fd "    PLB_Clk,"
    puts $fd "    PLB_Rst,"
    puts $fd "    PLB_abort,"
    puts $fd "    PLB_masterID,"
    puts $fd "    PLB_PAValid,"
    puts $fd "    PLB_RNW,"
    puts $fd "    PLB_ABus,"
    puts $fd "    PLB_wrDBus,"
    puts $fd "    Sl_MBusy,"
    if {$v4} {
	puts $fd "    Sl_MRdErr,"
	puts $fd "    Sl_MWrErr,"
    } else {
    puts $fd "    Sl_MErr,"
    }
    puts $fd "    Sl_rearbitrate,"
    puts $fd "    Sl_wait,"
    puts $fd "    Sl_SSize,"
    puts $fd "    Sl_addrAck,"
    puts $fd "    Sl_rdBTerm,"
    puts $fd "    Sl_rdComp,"
    puts $fd "    Sl_rdDAck,"
    puts $fd "    Sl_rdWdAddr,"
    puts $fd "    Sl_rdDBus,"
    puts $fd "    Sl_wrBTerm,"
    puts $fd "    Sl_wrComp,"
    puts $fd "    Sl_wrDAck,"
    if {$v4} {
	puts $fd "    Sl_MIRQ,"
    }
    puts $fd "    IP2Plb_data,"
    puts $fd "    Plb2IP_Reset,"
    puts $fd "    Plb2IP_Clk,"
    puts $fd "    Plb2IP_CS,"
    puts $fd "    Plb2IP_RdCE,"
    puts $fd "    Plb2IP_WrCE,"
    puts $fd "    Plb2IP_raddr,"
    puts $fd "    Plb2IP_waddr,"
    puts $fd "    Plb2IP_data"
    puts $fd "  );"
    puts $fd ""
}

proc genPlbConn {fd con no total rstName clkName v4} {
    set selWidth [calcAddrBits $total]
    set addrWidth [expr $selWidth + 4]
    set streamsel "([expr $addrWidth - 1] downto 4)"
    
    set type [lindex $con 0]
    set dir [lindex $con 1]
    set iname [lindex $con 2]
    set width [lindex $con 3]
    switch $type {
        {stream} {
	    set addr [makeVec $no $selWidth]
 	    puts $fd "  ${iname}_cs <="
	    puts $fd "    '1' when Plb2IP_raddr${streamsel} = $addr and Plb2IP_RdCE = '1' else"
	    puts $fd "    '1' when Plb2IP_waddr${streamsel} = $addr and Plb2IP_WrCE = '1' else"
	    puts $fd "    '0';\n"
	    if {$dir == "in"} {
	        puts $fd "  ${iname}_if: plb_to_stream"
	    } else {
	        puts $fd "  ${iname}_if: stream_to_plb"
	    }
	    puts $fd "    generic map ("
		 puts $fd "      C_SPLB_DWIDTH => C_SPLB_DWIDTH,"
	    puts $fd "      datawidth => $width"
	    puts $fd "    )"
	    puts $fd "    port map ("
	    puts $fd "      clk => Plb2IP_Clk,"
	    puts $fd "      plb_raddr => Plb2IP_raddr(3 downto 3),"
	    puts $fd "      plb_waddr => Plb2IP_waddr(3 downto 3),"
	    puts $fd "      plb_ce => ${iname}_cs,"
	    puts $fd "      plb_read => Plb2IP_RdCe,"
	    puts $fd "      plb_reset => Plb2IP_Reset,"
	    puts $fd "      plb_write => Plb2IP_WrCe,"
	    puts $fd "      plb_wdata => Plb2IP_Data,"
	    puts $fd "      plb_rdata => ${iname}_odata,"
	    puts $fd "      stream_en => ${iname}_en,"
	    puts $fd "      stream_rdy => ${iname}_rdy,"
	    puts $fd "      stream_eos => ${iname}_eos,"
	    puts $fd "      stream_data => ${iname}_idata"
	    puts $fd "    );\n"
        }
        {signal} {
	    set addr [makeVec $no $selWidth]
 	    puts $fd "  ${iname}_cs <="
	    puts $fd "    '1' when Plb2IP_raddr${streamsel} = $addr and Plb2IP_RdCE = '1' else"
	    puts $fd "    '1' when Plb2IP_waddr${streamsel} = $addr and Plb2IP_WrCE = '1' else"
	    puts $fd "    '0';\n"
	    if {$dir == "in"} {
	        puts $fd "  ${iname}_if: plb_to_signal"
	    } else {
	        puts $fd "  ${iname}_if: signal_to_plb"
	    }
	    puts $fd "    generic map ("
		 puts $fd "      C_SPLB_DWIDTH => C_SPLB_DWIDTH,"
	    if {$width > 0} {
		puts $fd "      datawidth => $width"
	    } else {
		puts $fd "      datawidth => 1"
	    }
	    puts $fd "    )"
	    puts $fd "    port map ("
	    puts $fd "      clk => Plb2IP_Clk,"
	    puts $fd "      plb_raddr => Plb2IP_raddr(3 downto 3),"
	    puts $fd "      plb_waddr => Plb2IP_waddr(3 downto 3),"
	    puts $fd "      plb_ce => ${iname}_cs,"
	    puts $fd "      plb_read => Plb2IP_RdCe,"
	    puts $fd "      plb_reset => Plb2IP_Reset,"
	    puts $fd "      plb_write => Plb2IP_WrCe,"
	    puts $fd "      plb_wdata => Plb2IP_Data,"
	    puts $fd "      plb_rdata => ${iname}_odata,"
	    puts $fd "      signal_en => ${iname}_en,"
	    if {$dir == "out"} {
	      puts $fd "      signal_rdy => ${iname}_rdy,"
            }
	    if {$width > 0} {
		puts $fd "      signal_data => ${iname}_idata"
	    } elseif {$dir == "in"} {
		puts $fd "      signal_data => open"
	    } else {
		puts $fd "      signal_data => \"0\""
	    }
	    puts $fd "    );\n"
        }
        {register} {
	    set addr [makeVec $no $selWidth]
 	    puts $fd "  ${iname}_cs <="
	    puts $fd "    '1' when Plb2IP_raddr${streamsel} = $addr and Plb2IP_RdCE = '1' else"
	    puts $fd "    '1' when Plb2IP_waddr${streamsel} = $addr and Plb2IP_WrCE = '1' else"
	    puts $fd "    '0';\n"
	    if {$dir == "in"} {
	        puts $fd "  ${iname}_if: plb_to_register"
	    } else {
	        puts $fd "  ${iname}_if: register_to_plb"
	    }
	    puts $fd "    generic map ("
		 puts $fd "      C_SPLB_DWIDTH => C_SPLB_DWIDTH,"
	    puts $fd "      datawidth => $width"
	    puts $fd "    )"
	    puts $fd "    port map ("
	    puts $fd "      clk => Plb2IP_Clk,"
	    puts $fd "      plb_raddr => Plb2IP_raddr(3 downto 3),"
	    puts $fd "      plb_waddr => Plb2IP_waddr(3 downto 3),"
	    puts $fd "      plb_ce => ${iname}_cs,"
	    puts $fd "      plb_read => Plb2IP_RdCe,"
	    puts $fd "      plb_reset => Plb2IP_Reset,"
	    puts $fd "      plb_write => Plb2IP_WrCe,"
	    puts $fd "      plb_wdata => Plb2IP_Data,"
	    puts $fd "      plb_rdata => ${iname}_odata,"
	    if {$dir == "in"} {
	      puts $fd "      register_en => ${iname}_en,"
	      puts $fd "      register_data => ${iname}_idata"
            } else {
	      puts $fd "      register_value => ${iname}_value"
            }
	    puts $fd "    );\n"
	}
	{memory} {
	    set iname [lindex $con 1]
	    puts $fd "  ${iname}_if: plb_dma"
	    puts $fd "    port map ("
	    puts $fd "      PLB_Rst => PLB_Rst,"
	    puts $fd "      PLB_Clk => PLB_Clk,"
	    puts $fd "      PLB_MAddrAck => PLB_MAddrAck,"
	    puts $fd "      PLB_MSSize => PLB_MSSize,"
	    puts $fd "      PLB_MRearbitrate => PLB_MRearbitrate,"
	    puts $fd "      PLB_MBusy => PLB_MBusy,"
	    if {$v4} {
		puts $fd "      PLB_MRdErr => PLB_MRdErr,"
		puts $fd "      PLB_MWrErr => PLB_MWrErr,"
	    } else {
	    puts $fd "      PLB_MErr => PLB_MErr,"
	    }
	    puts $fd "      PLB_MWrDAck => PLB_MWrDAck,"
	    puts $fd "      PLB_MRdDBus => PLB_MRdDBus,"
	    puts $fd "      PLB_MRdWdAddr => PLB_MRdWdAddr,"
	    puts $fd "      PLB_MRdDAck => PLB_MRdDAck,"
	    puts $fd "      PLB_MRdBTerm => PLB_MRdBTerm,"
	    puts $fd "      PLB_MWrBTerm => PLB_MWrBTerm,"
	    if {$v4} {
		puts $fd "      PLB_MTimeout => PLB_MTimeout,"
		puts $fd "      PLB_MIRQ => PLB_MIRQ,"
	    }
	    puts $fd "      M_request => M_request,"
	    puts $fd "      M_priority => M_priority,"
	    puts $fd "      M_busLock => M_busLock,"
	    puts $fd "      M_RNW => M_RNW,"
	    puts $fd "      M_BE => M_BE,"
	    puts $fd "      M_MSize => M_MSize,"
	    puts $fd "      M_size => M_size,"
	    puts $fd "      M_type => M_type,"
	    if {$v4} {
		puts $fd "      M_TAttribute => M_TAttribute,"
	    } else {
	    puts $fd "      M_compress => M_compress,"
	    puts $fd "      M_guarded => M_guarded,"
	    puts $fd "      M_ordered => M_ordered,"
	    }
	    puts $fd "      M_lockErr => M_lockErr,"
	    puts $fd "      M_abort => M_abort,"
	    puts $fd "      M_ABus => M_ABus,"
	    puts $fd "      M_wrDBus => M_wrDBus,"
	    puts $fd "      M_wrBurst => M_wrBurst,"
	    puts $fd "      M_rdBurst => M_rdBurst,"
	    if {$v4} {
		puts $fd "      M_UABus => M_UABus,"
	    }
	    puts $fd "      ICidata => IC${iname}_idata,"
	    puts $fd "      ICaddr => IC${iname}_addr,"
	    puts $fd "      ICnextaddr => IC${iname}_nextaddr,"
	    puts $fd "      ICwri => IC${iname}_wri,"
	    puts $fd "      ICre => IC${iname}_re,"
	    puts $fd "      ICodata => IC${iname}_odata,"
	    puts $fd "      ICreq => IC${iname}_req,"
	    puts $fd "      ICack => IC${iname}_ack,"
	    puts $fd "      ICblock => IC${iname}_block,"
	    puts $fd "      ICslice => IC${iname}_slice,"
	    puts $fd "      ICmode => IC${iname}_mode,"
	    puts $fd "      ICbase => IC${iname}_base,"
	    puts $fd "      ICsize => IC${iname}_size,"
	    puts $fd "      ICchunk => IC${iname}_chunk,"
	    puts $fd "      ICstride => IC${iname}_stride,"
	    puts $fd "      ICstart => IC${iname}_start,"
	    puts $fd "      ICcount => IC${iname}_count);\n"
	}
    }
}

proc genPlbMPD {fd multibus hasSlave hasMaster v4} {
    if {$v4} {
	set slaveGeneric "C_SPLB"
	set masterGeneric "C_MPLB"
    } else {
	set slaveGeneric "C_PLB"
	set masterGeneric "C_PLB"
    }
    puts $fd "# PLB interface"
    if {! $multibus} {
        puts $fd "PARAMETER c_baseaddr        = 0xFFFFFFFF, DT = std_logic_vector, BUS = SPLB, MIN_SIZE = 0x0f"
        puts $fd "PARAMETER c_highaddr        = 0x00000000, DT = std_logic_vector, BUS = SPLB"
    } else {
        puts $fd "PARAMETER c_plb_baseaddr        = 0xFFFFFFFF, DT = std_logic_vector, BUS = SPLB, MIN_SIZE = 0x0f"
        puts $fd "PARAMETER c_plb_highaddr        = 0x00000000, DT = std_logic_vector, BUS = SPLB"
    }
    puts $fd "PARAMETER ${slaveGeneric}_NUM_MASTERS = 8, DT = INTEGER, BUS = SPLB"
    puts $fd "PARAMETER ${slaveGeneric}_MID_WIDTH = 3, DT = INTEGER, BUS = SPLB"
    puts $fd "PARAMETER ${slaveGeneric}_AWIDTH      = 32,         DT = integer, BUS = SPLB"
    puts $fd "PARAMETER ${slaveGeneric}_DWIDTH      = 64,         DT = integer, BUS = SPLB"
    if {$hasMaster} {
	if {$v4} {
	    puts $fd "PARAMETER ${masterGeneric}_AWIDTH      = 32,         DT = integer, BUS = MPLB"
	    puts $fd "PARAMETER ${masterGeneric}_DWIDTH      = 64,         DT = integer, BUS = MPLB"
	}
    }

    puts $fd ""
    puts $fd "PORT PLB_Rst = PLB_Rst, DIR = I, SIGIS = Rst, BUS = SPLB"
    puts $fd "PORT PLB_abort = PLB_abort, DIR = I, BUS = SPLB"
    puts $fd "PORT PLB_Clk = \"\", DIR = I, SIGIS = Clk, BUS = SPLB"
    puts $fd "PORT PLB_ABus = PLB_ABus, DIR = I, VEC = \[0:(${slaveGeneric}_AWIDTH-1)\], BUS = SPLB"
    puts $fd "PORT PLB_BE = PLB_BE, DIR = I, VEC = \[0:((${slaveGeneric}_DWIDTH/8)-1)\], BUS = SPLB"
    puts $fd "PORT PLB_busLock = PLB_busLock, DIR = I, BUS = SPLB"
    if {$v4} {
	puts $fd "PORT PLB_TAttribute = PLB_TAttribute, DIR = I, VEC = \[0:15\], BUS = SPLB"
    } else {
    puts $fd "PORT PLB_compress = PLB_compress, DIR = I, BUS = SPLB"
    puts $fd "PORT PLB_guarded = PLB_guarded, DIR = I, BUS = SPLB"
    puts $fd "PORT PLB_ordered = PLB_ordered, DIR = I, BUS = SPLB"
    }
    puts $fd "PORT PLB_lockErr = PLB_lockErr, DIR = I, BUS = SPLB"
    puts $fd "PORT PLB_masterID = PLB_masterID, DIR = I, VEC = \[0:(${slaveGeneric}_MID_WIDTH-1)\], BUS = SPLB"
    puts $fd "PORT PLB_MSize = PLB_MSize, DIR = I, VEC = \[0:1\], BUS = SPLB"
    puts $fd "PORT PLB_PAValid = PLB_PAValid, DIR = I, BUS = SPLB"
    puts $fd "PORT PLB_RNW = PLB_RNW, DIR = I, BUS = SPLB"
    puts $fd "PORT PLB_size = PLB_size, DIR = I, VEC = \[0:3\], BUS = SPLB"
    puts $fd "PORT PLB_type = PLB_type, DIR = I, VEC = \[0:2\], BUS = SPLB"
    if {$v4} {
	puts $fd "PORT PLB_UABus = PLB_UABus, DIR = I, VEC = \[0:31\], BUS = SPLB"
    }
    puts $fd "PORT Sl_addrAck = Sl_addrAck, DIR = O, BUS = SPLB"
    puts $fd "PORT Sl_MBusy = Sl_MBusy, DIR = O, VEC = \[0:(${slaveGeneric}_NUM_MASTERS-1)\], BUS = SPLB"
    if {$v4} {
	puts $fd "PORT Sl_MRdErr = Sl_MRdErr, DIR = O, VEC = \[0:(${slaveGeneric}_NUM_MASTERS-1)\], BUS = SPLB"
	puts $fd "PORT Sl_MWrErr = Sl_MWrErr, DIR = O, VEC = \[0:(${slaveGeneric}_NUM_MASTERS-1)\], BUS = SPLB"
    } else {
	puts $fd "PORT Sl_MErr = Sl_MErr, DIR = O, VEC = \[0:(${slaveGeneric}_NUM_MASTERS-1)\], BUS = SPLB"
    }
    puts $fd "PORT Sl_rearbitrate = Sl_rearbitrate, DIR = O, BUS = SPLB"
    puts $fd "PORT Sl_SSize = Sl_SSize, DIR = O, VEC = \[0:1\], BUS = SPLB"
    puts $fd "PORT Sl_wait = Sl_wait, DIR = O, BUS = SPLB"
    if {$v4} {
	puts $fd "PORT Sl_MIRQ = Sl_MIRQ, DIR = O, VEC = \[0:(${slaveGeneric}_NUM_MASTERS-1)\], BUS = SPLB"
    }
    puts $fd "PORT PLB_rdPrim = PLB_rdPrim, DIR = I, BUS = SPLB"
    puts $fd "PORT PLB_SAValid = PLB_SAValid, DIR = I, BUS = SPLB"
    puts $fd "PORT PLB_wrPrim = PLB_wrPrim, DIR = I, BUS = SPLB"
    puts $fd "PORT PLB_wrBurst = PLB_wrBurst, DIR = I, BUS = SPLB"
    puts $fd "PORT PLB_wrDBus = PLB_wrDBus, DIR = I, VEC = \[0:(${slaveGeneric}_DWIDTH-1)\], BUS = SPLB"
    puts $fd "PORT Sl_wrBTerm = Sl_wrBTerm, DIR = O, BUS = SPLB"
    puts $fd "PORT Sl_wrComp = Sl_wrComp, DIR = O, BUS = SPLB"
    puts $fd "PORT Sl_wrDAck = Sl_wrDAck, DIR = O, BUS = SPLB"
    puts $fd "PORT PLB_rdBurst = PLB_rdBurst, DIR = I, BUS = SPLB"
    puts $fd "PORT Sl_rdBTerm = Sl_rdBTerm, DIR = O, BUS = SPLB"
    puts $fd "PORT Sl_rdComp = Sl_rdComp, DIR = O, BUS = SPLB"
    puts $fd "PORT Sl_rdDAck = Sl_rdDAck, DIR = O, BUS = SPLB"
    puts $fd "PORT Sl_rdDBus = Sl_rdDBus, DIR = O, VEC = \[0:(${slaveGeneric}_DWIDTH-1)\], BUS = SPLB"
    puts $fd "PORT Sl_rdWdAddr = Sl_rdWdAddr, DIR = O, VEC = \[0:3\], BUS = SPLB"
    if {$v4} {
	puts $fd "PORT PLB_rdPendReq = PLB_rdPendReq, DIR = I, BUS = SPLB"
	puts $fd "PORT PLB_wrPendReq = PLB_wrPendReq, DIR = I, BUS = SPLB"
	puts $fd "PORT PLB_rdPendPri = PLB_rdPendPri, DIR = I, VEC = \[0:1\], BUS = SPLB"
	puts $fd "PORT PLB_wrPendPri = PLB_wrPendPri, DIR = I, VEC = \[0:1\], BUS = SPLB"
    } else {
    puts $fd "PORT PLB_pendReq = PLB_pendReq, DIR = I, BUS = SPLB"
    puts $fd "PORT PLB_pendPri = PLB_pendPri, DIR = I, VEC = \[0:1\], BUS = SPLB"
    }
    puts $fd "PORT PLB_reqPri = PLB_reqPri, DIR = I, VEC = \[0:1\], BUS = SPLB"
    if {$hasMaster} {
	puts $fd "PORT PLB_MAddrAck=PLB_MAddrAck, DIR = I, BUS = MPLB"
	puts $fd "PORT PLB_MSSize=PLB_MSSize, DIR = I, VEC = \[0:1\], BUS=MPLB"
	puts $fd "PORT PLB_MRearbitrate=PLB_MRearbitrate, DIR = I, BUS = MPLB"
	puts $fd "PORT PLB_MBusy=PLB_MBusy, DIR = I, BUS = MPLB"
	if {$v4} {
	    puts $fd "PORT PLB_MRdErr=PLB_MRdErr, DIR = I, BUS = MPLB"
	    puts $fd "PORT PLB_MWrErr=PLB_MWrErr, DIR = I, BUS = MPLB"
	} else {
	puts $fd "PORT PLB_MErr=PLB_MErr, DIR = I, BUS = MPLB"
	}
	puts $fd "PORT PLB_MWrDAck=PLB_MWrDAck, DIR = I, BUS = MPLB"
	puts $fd "PORT PLB_MRdDBus=PLB_MRdDBus, DIR = I, VEC = \[0:(${masterGeneric}_DWIDTH-1)\], BUS=MPLB"
	puts $fd "PORT PLB_MRdWdAddr=PLB_MRdWdAddr, DIR = I, VEC = \[0:3\], BUS=MPLB"
	puts $fd "PORT PLB_MRdDAck=PLB_MRdDAck, DIR = I, BUS = MPLB"
	puts $fd "PORT PLB_MRdBTerm=PLB_MRdBTerm, DIR = I, BUS = MPLB"
	puts $fd "PORT PLB_MWrBTerm=PLB_MWrBTerm, DIR = I, BUS = MPLB"
	if {$v4} {
	    puts $fd "PORT PLB_MTimeout=PLB_MTimeout, DIR = I, BUS = MPLB"
	    puts $fd "PORT PLB_MIRQ=PLB_MIRQ, DIR = I, BUS = MPLB"
	}
	puts $fd "PORT M_abort=M_abort, DIR = O, BUS = MPLB"
	puts $fd "PORT M_ABus=M_ABus, DIR = O,  VEC = \[0:31\], BUS=MPLB"
	puts $fd "PORT M_BE=M_BE, DIR = O,  VEC = \[0:(${masterGeneric}_DWIDTH/8)-1\], BUS=MPLB"
	puts $fd "PORT M_busLock=M_busLock, DIR = O, BUS = MPLB"
	if {$v4} {
	    puts $fd "PORT M_TAttribute=M_TAttribute, DIR = O,  VEC = \[0:15\], BUS=MPLB"
	} else {
	puts $fd "PORT M_compress=M_compress, DIR = O, BUS = MPLB"
	puts $fd "PORT M_guarded=M_guarded, DIR = O, BUS = MPLB"
	    puts $fd "PORT M_ordered=M_ordered, DIR = O, BUS = MPLB"
	}
	puts $fd "PORT M_lockErr=M_lockErr, DIR = O, BUS = MPLB"
	puts $fd "PORT M_MSize=M_MSize, DIR = O, VEC = \[0:1\], BUS=MPLB"
	puts $fd "PORT M_priority=M_priority, DIR = O, VEC = \[0:1\], BUS=MPLB"
	puts $fd "PORT M_rdBurst=M_rdBurst, DIR = O, BUS = MPLB"
	puts $fd "PORT M_request=M_request, DIR = O, BUS = MPLB"
	puts $fd "PORT M_RNW=M_RNW, DIR = O, BUS = MPLB"
	puts $fd "PORT M_size=M_size, DIR = O, VEC = \[0:3\], BUS=MPLB"
	puts $fd "PORT M_type=M_type, DIR = O, VEC = \[0:2\], BUS=MPLB"
	puts $fd "PORT M_wrBurst=M_wrBurst, DIR = O, BUS = MPLB"
	puts $fd "PORT M_wrDBus=M_wrDBus, DIR = O, VEC = \[0:${masterGeneric}_DWIDTH-1\], BUS=MPLB"
	if {$v4} {
	    puts $fd "PORT M_UABus=M_UABus, DIR = O,  VEC = \[0:31\], BUS=MPLB"
	}
    }
    puts $fd ""
}

######################################################
# OPB SPECIFIC PROCS
######################################################

proc genOpbGenerics {fd multibus hasSlave hasMaster} {
    if {$hasSlave} {
        if {! $multibus} {
            puts $fd "    C_BASEADDR     : std_logic_vector(0 to 31) := X\"FFFFFFFF\";"
            puts $fd "    C_HIGHADDR     : std_logic_vector(0 to 31) := X\"00000000\";"
        } else {
            puts $fd "    C_OPB_BASEADDR     : std_logic_vector(0 to 31) := X\"FFFFFFFF\";"
            puts $fd "    C_OPB_HIGHADDR     : std_logic_vector(0 to 31) := X\"00000000\";"
        }
        puts $fd "    C_MIR_BASEADDR     : std_logic_vector(0 to 31) := X\"FFFFFFFF\";"
        puts $fd "    C_MIR_HIGHADDR     : std_logic_vector(0 to 31) := X\"00000000\";"
        puts $fd "    C_USER_ID_CODE     : integer                   := 3;"
    }
    puts $fd "    C_OPB_AWIDTH       : integer                   := 32;"
    puts -nonewline $fd "    C_OPB_DWIDTH       : integer                   := 32"
}

proc genOpbPorts {fd hasSlave hasMaster} {
    if {$hasSlave} {    
        puts $fd "    OPB_ABus     : in  std_logic_vector(0 to C_OPB_AWIDTH-1);"
        puts $fd "    OPB_BE       : in  std_logic_vector(0 to C_OPB_DWIDTH/8-1);"
        puts $fd "    OPB_Clk      : in  std_logic;"
        puts $fd "    OPB_DBus     : in  std_logic_vector(0 to C_OPB_DWIDTH-1);"
        puts $fd "    OPB_RNW      : in  std_logic;"
        puts $fd "    OPB_Rst      : in  std_logic;"
        puts $fd "    OPB_select   : in  std_logic;"
        puts -nonewline $fd "    OPB_seqAddr  : in  std_logic"
    } else {
        puts $fd "    OPB_Clk      : in  std_logic;"
        puts $fd "    OPB_Rst      : in  std_logic;"
        puts -nonewline $fd "    OPB_DBus     : in  std_logic_vector(0 to C_OPB_DWIDTH-1)"
    }
    if {$hasMaster} {
        puts $fd ";\n    OPB_MGrant   : in  std_logic := '0';"
        puts $fd "    OPB_xferAck  : in  std_logic := '0';"
        puts $fd "    OPB_errAck   : in  std_logic := '0';"
        puts $fd "    OPB_retry    : in  std_logic := '0';"
        puts -nonewline $fd "    OPB_timeout  : in  std_logic := '0'"
    }
    if {$hasSlave} {
        puts $fd ";\n    Sln_DBus     : out std_logic_vector(0 to C_OPB_DWIDTH-1);"
        puts $fd "    Sln_errAck   : out std_logic;"
        puts $fd "    Sln_retry    : out std_logic;"
        puts $fd "    Sln_toutSup  : out std_logic;"
        puts -nonewline $fd "    Sln_xferAck  : out std_logic"
    }
    if {$hasMaster} {
        puts $fd ";\n    Mn_ABus           : out std_logic_vector(0 to 31);"
        puts $fd "    Mn_DBus           : out std_logic_vector(0 to 31);"
        puts $fd "    Mn_request        : out std_logic;"
        puts $fd "    Mn_busLock        : out std_logic;"
        puts $fd "    Mn_select         : out std_logic;"
        puts $fd "    Mn_RNW            : out std_logic;"
        puts $fd "    Mn_BE             : out std_logic_vector(0 to 3);"
        puts -nonewline $fd "    Mn_seqAddr        : out std_logic"
    }
}

proc genOpbConnSignals {fd con} {
    set type [lindex $con 0]
    set iname [lindex $con 2]
    if {$type != "memory"} {
	puts $fd "  signal ${iname}_cs : std_logic;"
	puts $fd "  signal ${iname}_odata : std_logic_vector(31 downto 0);"
    }
}

proc genOpbSignals {fd noAttachments} {
    puts $fd "  signal Opb2IP_Addr        : std_logic_vector(0 to C_OPB_AWIDTH-1);"
    puts $fd "  signal Opb2IP_BE          : std_logic_vector(0 to C_OPB_DWIDTH/8-1);"
    puts $fd "  signal Opb2IP_Burst       : std_logic;"
    puts $fd "  signal Opb2IP_Clk         : std_logic;"
    puts $fd "  signal Opb2IP_CS          : std_logic;"
    puts $fd "  signal Opb2IP_Data        : std_logic_vector(0 to C_OPB_DWIDTH-1);"
    puts $fd "  signal Opb2IP_RdCE        : std_logic;"
    puts $fd "  signal Opb2IP_Reset       : std_logic;"
    puts $fd "  signal Opb2IP_RNW         : std_logic;"
    puts $fd "  signal Opb2IP_WrCE        : std_logic;"
    puts $fd "  signal IP2Opb_Ack         : std_logic;"
    puts $fd "  signal IP2Opb_Data        : std_logic_vector(0 to C_OPB_DWIDTH-1);"
    puts $fd "  signal IP2Opb_Error       : std_logic;"
    puts $fd "  signal IP2Opb_PostedWrInh : std_logic;"
    puts $fd "  signal IP2Opb_Retry       : std_logic;"
    puts $fd "  signal IP2Opb_ToutSup     : std_logic;"
    set addrWidth [calcAddrBits $noAttachments]
    if {$addrWidth > 0} {
	puts $fd "  signal opbsel : std_logic_vector([expr $addrWidth - 1] downto 0);"
    }

    puts $fd "\n  component opb_ipif_ssp0 is"
    puts $fd "    generic"
    puts $fd "    ("
    puts $fd "      C_BASEADDR         : std_logic_vector(0 to 31) := X\"FFFFFFFF\";"
    puts $fd "      C_HIGHADDR         : std_logic_vector(0 to 31) := X\"00000000\";"
    puts $fd "      C_MIR_BASEADDR     : std_logic_vector(0 to 31) := X\"FFFFFFFF\";"
    puts $fd "      C_MIR_HIGHADDR     : std_logic_vector(0 to 31) := X\"00000000\";"
    puts $fd "      C_USER_ID_CODE     : integer                   := 1;"
    puts $fd "      C_PIPELINE_MODEL   : integer                   := 4;"    
    puts $fd "      C_OPB_AWIDTH       : integer                   := 32;"
    puts $fd "      C_OPB_DWIDTH       : integer                   := 32;"    
    puts $fd "      C_FAMILY           : string                    := \"virtex2\""
    puts $fd "    );"
    puts $fd "    port"
    puts $fd "    ("
    puts $fd "        OPB_ABus           : in  std_logic_vector(0 to C_OPB_AWIDTH-1);"
    puts $fd "        OPB_BE             : in  std_logic_vector(0 to C_OPB_DWIDTH/8-1);"
    puts $fd "        OPB_Clk            : in  std_logic;"
    puts $fd "        OPB_DBus           : in  std_logic_vector(0 to C_OPB_DWIDTH-1);"
    puts $fd "        OPB_RNW            : in  std_logic;"
    puts $fd "        OPB_Rst            : in  std_logic;"
    puts $fd "        OPB_select         : in  std_logic;"
    puts $fd "        OPB_seqAddr        : in  std_logic;"
    puts $fd "        Sln_DBus           : out std_logic_vector(0 to C_OPB_DWIDTH-1);"
    puts $fd "        Sln_errAck         : out std_logic;"
    puts $fd "        Sln_retry          : out std_logic;"
    puts $fd "        Sln_toutSup        : out std_logic;"
    puts $fd "        Sln_xferAck        : out std_logic;"
    puts $fd "        Bus2IP_Addr        : out std_logic_vector(0 to C_OPB_AWIDTH-1);"
    puts $fd "        Bus2IP_BE          : out std_logic_vector(0 to C_OPB_DWIDTH/8-1);"
    puts $fd "        Bus2IP_Burst       : out std_logic;"
    puts $fd "        Bus2IP_Clk         : out std_logic;"
    puts $fd "        Bus2IP_CS          : out std_logic;"
    puts $fd "        Bus2IP_Data        : out std_logic_vector(0 to C_OPB_DWIDTH-1);"
    puts $fd "        Bus2IP_RdCE        : out std_logic;"
    puts $fd "        Bus2IP_Reset       : out std_logic;"
    puts $fd "        Bus2IP_RNW         : out std_logic;"
    puts $fd "        Bus2IP_WrCE        : out std_logic;"
    puts $fd "        IP2Bus_Ack         : in  std_logic;"
    puts $fd "        IP2Bus_Data        : in  std_logic_vector(0 to C_OPB_DWIDTH-1);"
    puts $fd "        IP2Bus_Error       : in  std_logic;"
    puts $fd "        IP2Bus_PostedWrInh : in  std_logic;"
    puts $fd "        IP2Bus_Retry       : in  std_logic;"
    puts $fd "        IP2Bus_ToutSup     : in  std_logic"
    puts $fd "    );"
    puts $fd "  end component opb_ipif_ssp0;"
}

proc genOpbIpif {fd multibus} {
    puts $fd "  OPB_IPIF_SSP0_I : opb_ipif_ssp0"
    puts $fd "    generic map"
    puts $fd "    ("
    if {$multibus} {
        puts $fd "      C_BASEADDR         => C_OPB_BASEADDR,"
        puts $fd "      C_HIGHADDR         => C_OPB_HIGHADDR,"
    } else {
        puts $fd "      C_BASEADDR         => C_BASEADDR,"
        puts $fd "      C_HIGHADDR         => C_HIGHADDR,"
    }
    puts $fd "      C_MIR_BASEADDR     => C_MIR_BASEADDR,"
    puts $fd "      C_MIR_HIGHADDR     => C_MIR_HIGHADDR,"
    puts $fd "      C_USER_ID_CODE     => C_USER_ID_CODE,"
    puts $fd "      C_PIPELINE_MODEL   => 5,"
    puts $fd "      C_OPB_AWIDTH       => C_OPB_AWIDTH,"
    puts $fd "      C_OPB_DWIDTH       => C_OPB_DWIDTH,"
    puts $fd "      C_FAMILY           => C_FAMILY"
    puts $fd "    )"
    puts $fd "    port map"
    puts $fd "    ("
    puts $fd "      OPB_ABus           => OPB_ABus,"
    puts $fd "      OPB_BE             => OPB_BE,"
    puts $fd "      OPB_Clk            => OPB_Clk,"
    puts $fd "      OPB_DBus           => OPB_DBus,"
    puts $fd "      OPB_RNW            => OPB_RNW,"
    puts $fd "      OPB_Rst            => OPB_Rst,"
    puts $fd "      OPB_select         => OPB_select,"
    puts $fd "      OPB_seqAddr        => OPB_seqAddr,"
    puts $fd "      Sln_DBus           => Sln_DBus,"
    puts $fd "      Sln_errAck         => Sln_errAck,"
    puts $fd "      Sln_retry          => Sln_retry,"
    puts $fd "      Sln_toutSup        => Sln_toutSup,"
    puts $fd "      Sln_xferAck        => Sln_xferAck,"
    puts $fd "      Bus2IP_Addr        => Opb2IP_Addr,"
    puts $fd "      Bus2IP_BE          => Opb2IP_BE,"
    puts $fd "      Bus2IP_Burst       => Opb2IP_Burst,"
    puts $fd "      Bus2IP_Clk         => Opb2IP_Clk,"
    puts $fd "      Bus2IP_CS          => Opb2IP_CS,"
    puts $fd "      Bus2IP_Data        => Opb2IP_Data,"
    puts $fd "      Bus2IP_RdCE        => Opb2IP_RdCE,"
    puts $fd "      Bus2IP_Reset       => Opb2IP_Reset,"
    puts $fd "      Bus2IP_RNW         => Opb2IP_RNW,"
    puts $fd "      Bus2IP_WrCE        => Opb2IP_WrCE,"
    puts $fd "      IP2Bus_Ack         => IP2Opb_Ack,"
    puts $fd "      IP2Bus_Data        => IP2Opb_Data,"
    puts $fd "      IP2Bus_Error       => IP2Opb_Error,"
    puts $fd "      IP2Bus_PostedWrInh => IP2Opb_PostedWrInh,"
    puts $fd "      IP2Bus_Retry       => IP2Opb_Retry,"
    puts $fd "      IP2Bus_ToutSup     => IP2Opb_ToutSup"
    puts $fd "    );"
    puts $fd ""
    puts $fd "  IP2Opb_Ack         <= Opb2IP_WrCE or Opb2IP_RdCE;"
    puts $fd "  IP2Opb_Retry       <= '0';"
    puts $fd "  IP2Opb_Error       <= '0';"
    puts $fd "  IP2Opb_ToutSup     <= '0';"
    puts $fd "  IP2Opb_PostedWrInh <= '0';\n"
}

proc genOpbConn {fd con no total rstName clkName} {
    set addrWidth [calcAddrBits $total]
	set type [lindex $con 0]
	set dir [lindex $con 1]
	set iname [lindex $con 2]
	set width [lindex $con 3]
	switch $type {
	    {stream} {
		set addr [makeVec $no $addrWidth]
  		puts $fd "  ${iname}_cs <= Opb2IP_CS when opbsel = $addr else '0';\n"
		if {$dir == "in"} {
		    puts $fd "  ${iname}_if: opb_to_stream"
		} else {
		    puts $fd "  ${iname}_if: stream_to_opb"
		}
		puts $fd "    generic map ("
		puts $fd "      datawidth => $width"
		puts $fd "    )"
		puts $fd "    port map ("
		puts $fd "      clk => Opb2IP_Clk,"
		puts $fd "      opb_addr => Opb2IP_Addr(28 to 29),"
		puts $fd "      opb_ce => ${iname}_cs,"
		puts $fd "      opb_read => Opb2IP_RdCe,"
		puts $fd "      opb_reset => Opb2IP_Reset,"
		puts $fd "      opb_write => Opb2IP_WrCe,"
		puts $fd "      opb_wdata => Opb2IP_Data,"
		puts $fd "      opb_rdata => ${iname}_odata,"
		puts $fd "      stream_en => ${iname}_en,"
		puts $fd "      stream_rdy => ${iname}_rdy,"
		puts $fd "      stream_eos => ${iname}_eos,"
		puts $fd "      stream_data => ${iname}_idata"
		puts $fd "    );\n"
	        incr no
	    }
	    {signal} {
		set addr [makeVec $no $addrWidth]
  		puts $fd "  ${iname}_cs <= Opb2IP_CS when opbsel = $addr else '0';\n"
		if {$dir == "in"} {
		    puts $fd "  ${iname}_if: opb_to_signal"
		} else {
		    puts $fd "  ${iname}_if: signal_to_opb"
		}
		puts $fd "    generic map ("
		if {$width > 0} {
		    puts $fd "      datawidth => $width"
		} else {
		    puts $fd "      datawidth => 1"
		}
		puts $fd "    )"
		puts $fd "    port map ("
		puts $fd "      clk => Opb2IP_Clk,"
		puts $fd "      opb_addr => Opb2IP_Addr(28 to 29),"
		puts $fd "      opb_ce => ${iname}_cs,"
		puts $fd "      opb_read => Opb2IP_RdCe,"
		puts $fd "      opb_reset => Opb2IP_Reset,"
		puts $fd "      opb_write => Opb2IP_WrCe,"
		puts $fd "      opb_wdata => Opb2IP_Data,"
		puts $fd "      opb_rdata => ${iname}_odata,"
		puts $fd "      signal_en => ${iname}_en,"
		if {$dir == "out"} {
		  puts $fd "      signal_rdy => ${iname}_rdy,"
                }
		if {$width > 0} {
		    puts $fd "      signal_data => ${iname}_idata"
		} elseif {$dir == "in"} {
		    puts $fd "      signal_data => open"
		} else {
		    puts $fd "      signal_data => \"0\""
		}
		puts $fd "    );\n"
	        incr no
	    }
	    {register} {
		set addr [makeVec $no $addrWidth]
  		puts $fd "  ${iname}_cs <= Opb2IP_CS when opbsel = $addr else '0';\n"
		if {$dir == "in"} {
		    puts $fd "  ${iname}_if: opb_to_register"
		} else {
		    puts $fd "  ${iname}_if: register_to_opb"
		}
		puts $fd "    generic map ("
		puts $fd "      datawidth => $width"
		puts $fd "    )"
		puts $fd "    port map ("
		puts $fd "      clk => Opb2IP_Clk,"
		puts $fd "      opb_addr => Opb2IP_Addr(28 to 29),"
		puts $fd "      opb_ce => ${iname}_cs,"
		puts $fd "      opb_read => Opb2IP_RdCe,"
		puts $fd "      opb_reset => Opb2IP_Reset,"
		puts $fd "      opb_write => Opb2IP_WrCe,"
		puts $fd "      opb_wdata => Opb2IP_Data,"
		puts $fd "      opb_rdata => ${iname}_odata,"
		if {$dir == "in"} {
		  puts $fd "      register_en => ${iname}_en,"
		  puts $fd "      register_data => ${iname}_idata"
                } else {
		  puts $fd "      register_value => ${iname}_value"
                }
		puts $fd "    );\n"
	        incr no
	    }
	    {memory} {
		set iname [lindex $con 1]
		puts $fd "  ${iname}_if: opb_dma"
		puts $fd "    port map ("
		puts $fd "      reset => OPB_Rst, sclk => OPB_Clk, clk => OPB_Clk,"
		puts $fd "      Mn_ABus => Mn_ABus,"
		puts $fd "      Mn_DBus => Mn_DBus,"
		puts $fd "      Mn_request => Mn_request,"

		puts $fd "      Mn_busLock => Mn_busLock,"
		puts $fd "      Mn_select => Mn_select,"
		puts $fd "      Mn_RNW => Mn_RNW,"
		puts $fd "      Mn_BE => Mn_BE,"
		puts $fd "      Mn_seqAddr => Mn_seqAddr,"
		puts $fd "      OPB_DBus => OPB_DBus,"
		puts $fd "      OPB_MGrant => OPB_MGrant,"
		puts $fd "      OPB_xferAck => OPB_xferAck,"
		puts $fd "      OPB_errAck => OPB_errAck,"
		puts $fd "      OPB_retry => OPB_retry,"
		puts $fd "      OPB_timeout => OPB_timeout,"
		puts $fd "      ICidata => IC${iname}_idata,"
		puts $fd "      ICaddr => IC${iname}_addr,"
		puts $fd "      ICnextaddr => IC${iname}_nextaddr,"
		puts $fd "      ICwri => IC${iname}_wri,"
		puts $fd "      ICodata => IC${iname}_odata,"
		puts $fd "      ICack => IC${iname}_ack,"
		puts $fd "      ICreq => IC${iname}_req,"
		puts $fd "      ICblock => IC${iname}_block,"
		puts $fd "      ICmode => IC${iname}_mode,"
		puts $fd "      ICbase => IC${iname}_base,"
		puts $fd "      ICsize => IC${iname}_size,"
		puts $fd "      ICcount => IC${iname}_count);\n"
            }
	}
}

proc genOpbMPD {fd multibus hasSlave hasMaster} {
    puts $fd "# OPB interface"
    if {$hasSlave} {
        if {! $multibus} {
            puts $fd "PARAMETER c_baseaddr        = 0xFFFFFFFF, DT = std_logic_vector, BUS = SOPB, MIN_SIZE = 0x04"
            puts $fd "PARAMETER c_highaddr        = 0x00000000, DT = std_logic_vector, BUS = SOPB"
        } else {
            puts $fd "PARAMETER c_opb_baseaddr        = 0xFFFFFFFF, DT = std_logic_vector, BUS = SOPB, MIN_SIZE = 0x04"
            puts $fd "PARAMETER c_opb_highaddr        = 0x00000000, DT = std_logic_vector, BUS = SOPB"
        }
        puts $fd "PARAMETER c_mir_baseaddr    = 0xFFFFFFFF, DT = std_logic_vector, BUS = SOPB, MIN_SIZE = 0x04"
        puts $fd "PARAMETER c_mir_highaddr    = 0x00000000, DT = std_logic_vector, BUS = SOPB"
        puts $fd "PARAMETER c_user_id_code    = 3,          DT = integer"
    }
    puts $fd "PARAMETER c_opb_awidth      = 32,         DT = integer"
    puts $fd "PARAMETER c_opb_dwidth      = 32,         DT = integer"
    puts $fd ""
    if {$hasSlave} {
        puts $fd "PORT opb_abus    = OPB_ABus,    DIR = IN, VEC = \[0:(c_opb_awidth-1)\],     BUS = SOPB"
        puts $fd "PORT opb_be      = OPB_BE,      DIR = IN, VEC = \[0:((c_opb_dwidth/8)-1)\], BUS = SOPB"
        puts $fd "PORT opb_clk     = \"\",          DIR = IN,                                 BUS = SOPB"
        puts $fd "PORT opb_dbus    = OPB_DBus,    DIR = IN, VEC = \[0:(c_opb_dwidth-1)\],     BUS = SOPB"
        puts $fd "PORT opb_rnw     = OPB_RNW,     DIR = IN,                                 BUS = SOPB"
        puts $fd "PORT opb_rst     = OPB_Rst,     DIR = IN,                                 BUS = SOPB"
        puts $fd "PORT opb_select  = OPB_select,  DIR = IN,                                 BUS = SOPB"
        puts $fd "PORT opb_seqaddr = OPB_seqAddr, DIR = IN,                                 BUS = SOPB"
        puts $fd "PORT sln_dbus    = Sl_DBus,     DIR = OUT, VEC = \[0:(c_opb_dwidth-1)\],    BUS = SOPB"
        puts $fd "PORT sln_errack  = Sl_errAck,   DIR = OUT,                                BUS = SOPB"
        puts $fd "PORT sln_retry   = Sl_retry,    DIR = OUT,                                BUS = SOPB"
        puts $fd "PORT sln_toutsup = Sl_toutSup,  DIR = OUT,                                BUS = SOPB"
        puts $fd "PORT sln_xferack = Sl_xferAck,  DIR = OUT,                                BUS = SOPB"
    } else {
        puts $fd "PORT opb_clk     = \"\",          DIR = IN,                                 BUS = MOPB"
        puts $fd "PORT opb_rst     = OPB_Rst,     DIR = IN,                                 BUS = MOPB"
        puts $fd "PORT opb_dbus    = OPB_DBus,    DIR = IN, VEC = \[0:(c_opb_dwidth-1)\],     BUS = MOPB"
    }
    if {$hasMaster} {
        puts $fd "PORT mn_abus     = M_ABus,      DIR = OUT, VEC = \[0:(c_opb_dwidth-1)\],     BUS = MOPB"
        puts $fd "PORT mn_dbus     = M_DBus,      DIR = OUT, VEC = \[0:(c_opb_dwidth-1)\],     BUS = MOPB"
        puts $fd "PORT mn_request  = M_request,   DIR = OUT,                                 BUS = MOPB"
        puts $fd "PORT mn_buslock  = M_busLock,   DIR = OUT,                                 BUS = MOPB"
        puts $fd "PORT mn_select   = M_select,    DIR = OUT,                                 BUS = MOPB"
        puts $fd "PORT mn_rnw      = M_RNW,       DIR = OUT,                                 BUS = MOPB"
        puts $fd "PORT mn_be       = M_BE,        DIR = OUT, VEC = \[0:((c_opb_dwidth/8)-1)\], BUS = MOPB"
        puts $fd "PORT mn_seqaddr  = M_seqAddr,   DIR = OUT,                                 BUS = MOPB"
        puts $fd "PORT opb_mgrant  = OPB_MGrant,  DIR = IN,                                  BUS = MOPB"
        puts $fd "PORT opb_xferack = OPB_xferAck, DIR = IN,                                  BUS = MOPB"
        puts $fd "PORT opb_errAck  = OPB_errAck,  DIR = IN,                                  BUS = MOPB"
        puts $fd "PORT opb_retry   = OPB_retry,   DIR = IN,                                  BUS = MOPB"
        puts $fd "PORT opb_timeout = OPB_timeout, DIR = IN,                                  BUS = MOPB"
    }
    puts $fd ""
}

######################################################
# FSL SPECIFIC PROCS
######################################################

proc genFslPorts {fd no dir} {
	if {$no == 0} {
		puts $fd "    FSL_Clk : in std_logic;"
		puts $fd "    FSL_Rst : in std_logic;"
	}
	if {$dir == "in"} {
		puts $fd "    FSL${no}_S_Clk : out std_logic;"
		puts $fd "    FSL${no}_S_Exists : in std_logic;"
		puts $fd "    FSL${no}_S_Data : in std_logic_vector (31 downto 0);"
		puts $fd "    FSL${no}_S_Control : in std_logic;"

		puts -nonewline $fd "    FSL${no}_S_Read : out std_logic"
	} else {
		puts $fd "    FSL${no}_M_Full : in std_logic;"
		puts $fd "    FSL${no}_M_Clk : out std_logic;"
		puts $fd "    FSL${no}_M_Data : out std_logic_vector (31 downto 0);"
		puts $fd "    FSL${no}_M_Control : out std_logic;"
		puts -nonewline $fd "    FSL${no}_M_Write : out std_logic"
	}
}

proc genFslConn {fd con no rstName clkName} {
	set type [lindex $con 0]
	set dir [lindex $con 1]
	set iname [lindex $con 2]
	set width [lindex $con 3]
	switch $type {
		{stream} {
			if {$dir == "in"} {
				puts $fd "  FSL${no}_S_Clk <= $clkName;"
				puts $fd "  ${iname}_if: fsl_to_stream"
				puts $fd "    generic map ("
				puts $fd "      datawidth => $width"
				puts $fd "    )"
				puts $fd "    port map ("
				puts $fd "      $rstName, $clkName,"
				puts $fd "      FSL${no}_S_Exists,"
				puts $fd "      FSL${no}_S_Data,"
				puts $fd "      FSL${no}_S_Control,"
				puts $fd "      FSL${no}_S_Read,"
				puts $fd "      stream_en => ${iname}_en,"
				puts $fd "      stream_rdy => ${iname}_rdy,"
				puts $fd "      stream_eos => ${iname}_eos,"
				puts $fd "      stream_data => ${iname}_idata"
				puts $fd "    );\n"
			} else {
				puts $fd "  FSL${no}_M_Clk <= $clkName;"
				puts $fd "  ${iname}_if: stream_to_fsl"
				puts $fd "    generic map ("
				puts $fd "      datawidth => $width"
				puts $fd "    )"
				puts $fd "    port map ("
				puts $fd "      $rstName, $clkName,"
				puts $fd "      FSL${no}_M_Full,"
				puts $fd "      FSL${no}_M_Write,"
				puts $fd "      FSL${no}_M_Control,"
				puts $fd "      FSL${no}_M_Data,"
				puts $fd "      stream_en => ${iname}_en,"
				puts $fd "      stream_rdy => ${iname}_rdy,"
				puts $fd "      stream_eos => ${iname}_eos,"
				puts $fd "      stream_data => ${iname}_idata"
				puts $fd "    );\n"
			}
		}
	}
}

proc genFslMPD {fd connections} {
	puts $fd "# FSL interface"
	
	set buslist [list]
	set no 0
	foreach con $connections {
		set busname [selectBus $con]
		if {$busname == "fsl"} {
			set dir [lindex $con 1]
			if {$dir == "in"} {
				puts $fd "PORT FSL${no}_S_Clk = FSL_S_Clk, DIR=out, BUS=SFSL$no"
				puts $fd "PORT FSL${no}_S_Data = FSL_S_Data, DIR=in, VEC=\[31:0\], BUS=SFSL$no"
				puts $fd "PORT FSL${no}_S_Control = FSL_S_Control, DIR=in, BUS=SFSL$no"
				puts $fd "PORT FSL${no}_S_Read = FSL_S_Read, DIR=out, BUS=SFSL$no"
				puts $fd "PORT FSL${no}_S_Exists = FSL_S_Exists, DIR=in, BUS=SFSL$no"
				lappend buslist SFSL$no
			} else {
				puts $fd "PORT FSL${no}_M_Clk = FSL_M_Clk, DIR=out, BUS=MFSL$no"
				puts $fd "PORT FSL${no}_M_Data = FSL_M_Data, DIR=out, VEC=\[31:0\], BUS=MFSL$no"
				puts $fd "PORT FSL${no}_M_Control = FSL_M_Control, DIR=out, BUS=MFSL$no"
				puts $fd "PORT FSL${no}_M_Write = FSL_M_Write, DIR=out, BUS=MFSL$no"
				puts $fd "PORT FSL${no}_M_Full = FSL_M_Full, DIR=in, BUS=MFSL$no"
				lappend buslist MFSL$no
			}
			incr no
		}
	}
	puts $fd "PORT FSL_Clk = \"\", DIR = IN, BUS = [join $buslist :]"
	puts $fd "PORT FSL_Rst = \"\", DIR = IN"
	puts $fd ""
}

######################################################
# DVI SPECIFIC PROCS
######################################################

proc genDVIPorts {fd no dir} {
	if {$no == 0} {
		puts $fd "    DVI_Clk : in std_ulogic;"
		puts $fd "    DVI_Rst : in std_ulogic;"
	}
	if {$dir == "in"} {
		puts $fd "    DVI_in_vsync : in std_ulogic;"
		puts $fd "    DVI_in_hsync : in std_ulogic;"
		puts $fd "    DVI_in_de : in std_ulogic;"
		puts $fd "    DVI_in_r : in std_ulogic_vector (7 downto 0);"
		puts $fd "    DVI_in_g : in std_ulogic_vector (7 downto 0);"
		puts -nonewline $fd "    DVI_in_b : in std_ulogic_vector (7 downto 0)"
	} else {
		puts $fd "    DVI_out_vsync : out std_ulogic;"
		puts $fd "    DVI_out_hsync : out std_ulogic;"
		puts $fd "    DVI_out_de : out std_ulogic;"
		puts $fd "    DVI_out_r : out std_ulogic_vector (7 downto 0);"
		puts $fd "    DVI_out_g : out std_ulogic_vector (7 downto 0);"
		puts -nonewline $fd "    DVI_out_b : out std_ulogic_vector (7 downto 0)"
	}
}

proc genDVIConn {fd con no rstName clkName} {
	set type [lindex $con 0]
	set dir [lindex $con 1]
	set iname [lindex $con 2]
	set width [lindex $con 3]
	switch $type {
		{stream} {
			if {$dir == "in"} {
				puts $fd "  ${iname}_if: DVI_to_stream"
				puts $fd "    generic map ("
				puts $fd "      datawidth => $width"
				puts $fd "    )"
				puts $fd "    port map ("
				puts $fd "      $rstName, $clkName,"
				puts $fd "      DVI_in_vsync,"
				puts $fd "      DVI_in_hsync,"
				puts $fd "      DVI_in_de,"
				puts $fd "      DVI_in_r,"
				puts $fd "      DVI_in_g,"
				puts $fd "      DVI_in_b,"
				puts $fd "      stream_en => ${iname}_en,"
				puts $fd "      stream_rdy => ${iname}_rdy,"
				puts $fd "      stream_eos => ${iname}_eos,"
				puts $fd "      stream_data => ${iname}_idata"
				puts $fd "    );\n"
			} else {
				puts $fd "  ${iname}_if: stream_to_DVI"
				puts $fd "    generic map ("
				puts $fd "      datawidth => $width"
				puts $fd "    )"
				puts $fd "    port map ("
				puts $fd "      $rstName, $clkName,"
				puts $fd "      DVI_out_vsync,"
				puts $fd "      DVI_out_hsync,"
				puts $fd "      DVI_out_de,"
				puts $fd "      DVI_out_r,"
				puts $fd "      DVI_out_g,"
				puts $fd "      DVI_out_b,"
				puts $fd "      stream_en => ${iname}_en,"
				puts $fd "      stream_rdy => ${iname}_rdy,"
				puts $fd "      stream_eos => ${iname}_eos,"
				puts $fd "      stream_data => ${iname}_idata"
				puts $fd "    );\n"
			}
		}
	}
}

proc genDVIMPD {fd connections} {
	puts $fd "# DVI interface"
	
	set buslist [list]
	set no 0
	foreach con $connections {
		set busname [selectBus $con]
		if {$busname == "dvi"} {
			set dir [lindex $con 1]
			if {$dir == "in"} {
				puts $fd "PORT DVI_in_vsync = vsync, DIR=I, BUS = IMPULSE_DVI_IN"
				puts $fd "PORT DVI_in_hsync = hsync, DIR=I, BUS = IMPULSE_DVI_IN"
				puts $fd "PORT DVI_in_de = de, DIR=I, BUS = IMPULSE_DVI_IN"
				puts $fd "PORT DVI_in_r = red, DIR=I, VEC=\[0\:(8-1)\], BUS = IMPULSE_DVI_IN"
				puts $fd "PORT DVI_in_g = green, DIR=I, VEC=\[0\:(8-1)\], BUS = IMPULSE_DVI_IN"
				puts $fd "PORT DVI_in_b = blue, DIR=I, VEC=\[0\:(8-1)\], BUS = IMPULSE_DVI_IN"
				puts $fd ""
				lappend buslist IMPULSE_DVI_IN
			} else {
				puts $fd "PORT DVI_out_vsync = vsync, DIR=O, BUS = IMPULSE_DVI_OUT"
				puts $fd "PORT DVI_out_hsync = hsync, DIR=O, BUS = IMPULSE_DVI_OUT"
				puts $fd "PORT DVI_out_de = de, DIR=O, BUS = IMPULSE_DVI_OUT"
				puts $fd "PORT DVI_out_r = red, DIR=O, VEC=\[0\:(8-1)\], BUS = IMPULSE_DVI_OUT"
				puts $fd "PORT DVI_out_g = green, DIR=O, VEC=\[0\:(8-1)\], BUS = IMPULSE_DVI_OUT"
				puts $fd "PORT DVI_out_b = blue, DIR=O, VEC=\[0\:(8-1)\], BUS = IMPULSE_DVI_OUT"
				puts $fd ""
				lappend buslist IMPULSE_DVI_OUT
			}
			incr no
		}
	}
	puts $fd "PORT DVI_Clk = \"\", DIR = IN"
	puts $fd "PORT DVI_Rst = \"\", DIR = IN"
	puts $fd ""
}

######################################################
# PCORE GENERATION
######################################################

proc genOutputMux {fd busName selName selWidth outName connections} {
	set count 0
	set signals [list]
	foreach con $connections {
		if {[selectBus $con] == $busName} {
			set type [lindex $con 0]
			switch $type {
				{stream} -
				{signal} -
				{register} {
					set iname [lindex $con 2]
					lappend signals ${iname}_odata
					incr count
				}
			}
		}
	}
    
	if {$count == 0} {
		puts $fd "  $outName <= [lindex $signals 0];"
		return
	}

	lappend signals $selName
	puts $fd "  ${busName}_mux: process ([join $signals {,}]) is"
	puts $fd "  begin"
	puts $fd "    case $selName is"
	set no 0
	foreach con $connections {
		if {[selectBus $con] == $busName} {
			set type [lindex $con 0]
			set iname [lindex $con 2]
			switch $type {
				{stream} -
				{signal} -
				{register} {
					set addr [makeVec $no $selWidth]
					puts $fd "      when $addr  => $outName <= ${iname}_odata;"
					incr no
				}
			}
		}
	}

	puts $fd "      when others => $outName <= (others => 'X');"
	puts $fd "    end case;"
	puts $fd "  end process ${busName}_mux;\n"
}

proc GenerateTop {outdir topname name connections slaves masters} {
        global _flags

	file mkdir [file join $outdir "pcore/"]

	set outfile [file join $outdir "pcore/${topname}.vhd"]
	if {[catch {set fd [open $outfile w]}]} {
		puts "unable to create $outfile"
		exit 1
	}

	gen_copyright_notice $fd
	
	# count attachments to each bus
	set attachments(apu440)    	0
	set attachments(apu)    	0
	set attachments(pcmcia) 	0
	set attachments(opb)    	0
	set attachments(plb)    	0
	set attachments(plb4)    	0
	set attachments(fsl)    	0
	set attachments(dvi)    	0
	set attachments(port)		0
	set attachments(generic)	0
	foreach con $connections {
		if {[lindex $con 0] != "memory"} {
			set busname [selectBus $con]
			incr attachments($busname)
		}
	}

 	gen_use_std_pkgs $fd 
	if {[lmember $slaves apu440]} {
		# print next line immediately after above call to 'gen_use_std_pkgs' until
		#  we figure out how to add 'std_logic_arith' to the standard package list
		puts $fd "use ieee.std_logic_arith.all;"
	}
	
 	set impulsePkgs [list]
	if {[lmember $slaves apu440]} {
		lappend impulsePkgs "apu_if"
	}
	if {[lmember $slaves apu]} {
		lappend impulsePkgs "apu_if"
	}
	if {[lmember $slaves pcmcia] || [lmember $masters pcmcia]} {
		#lappend impulsePkgs "pcmcia_if"
	}
	if {[lmember $slaves fsl]} {
		lappend impulsePkgs "fsl_if"
	}
	if {[lmember $slaves dvi]} {
		lappend impulsePkgs "dvi_if"
	}
	if {[lmember $slaves opb]} {
		gen_use_library_pkgs $fd "opb_ipif_ssp0_v1_00_b" [list]
	}
	if {[lmember $slaves opb] || [lmember $masters opb]} {
		lappend impulsePkgs "opb_if"
	}
	if {[lmember $slaves plb] || [lmember $masters plb]} {
		lappend impulsePkgs "plb_if"
	} elseif {[lmember $slaves plb4] || [lmember $masters plb4]} {
		lappend impulsePkgs "plb_if"
	}
 	gen_use_library_pkgs $fd "impulse" $impulsePkgs
 	gen_use_library_pkgs $fd "${topname}_v1_00_a" [list]

	if {[lmember $slaves plb] && [lmember $slaves opb]} {
		set multibus 1
	} elseif {[lmember $slaves plb4] && [lmember $slaves opb]} {
		set multibus 1
	} else {
		set multibus 0
	}

	set rstName ""
	set clkName ""

	puts $fd ""
	puts $fd "entity ${topname} is"
	puts $fd "  generic ("
	if {[lmember $slaves apu440]} {
	# APU440 is only supported on Virtex5
		puts -nonewline $fd "    C_FAMILY           : string                    := \"virtex5fx\""
	} elseif {[lmember $slaves apu] || [lmember $slaves pcmcia]} {
	# Pico's PCMCIA-based cards are Virtex4-only
	# APU is only supported on Virtex4
		puts -nonewline $fd "    C_FAMILY           : string                    := \"virtex4\""
	} else {
		puts -nonewline $fd "    C_FAMILY           : string                    := \"virtex2\""
	}
	# generate generics for each bus
	if {[lmember $slaves apu440]} {
#        puts $fd ";"
		genApu440Generics $fd $multibus [lmember $slaves apu440] [lmember $masters apu440]
	}
	if {[lmember $slaves apu]} {
#        puts $fd ";"
		genApuGenerics $fd $multibus [lmember $slaves apu] [lmember $masters apu]
	}
	if {[lmember $slaves pcmcia] || [lmember $masters pcmcia]} {
		puts $fd ";"
		genPcmciaGenerics $fd $multibus [lmember $slaves pcmcia] [lmember $masters pcmcia]
	}
	if {[lmember $slaves opb] || [lmember $masters opb]} {
		puts $fd ";"
		genOpbGenerics $fd $multibus [lmember $slaves opb] [lmember $masters opb]
	}
	if {[lmember $slaves plb] || [lmember $masters plb]} {
		puts $fd ";"
		genPlbGenerics $fd $multibus [lmember $slaves plb] [lmember $masters plb] 0
	} elseif {[lmember $slaves plb4] || [lmember $masters plb4]} {
		puts $fd ";"
		genPlbGenerics $fd $multibus [lmember $slaves plb4] [lmember $masters plb4] 1
	}
	puts $fd "  );"
	puts $fd "  port ("
	# generate ports for each slave bus
	set first 1
	if {[lmember $slaves apu440]} {
		if {!$first} {
			puts $fd ";"
		} else {
			set first 0
		}
		puts $fd "    -- apu ports"
		genApu440Ports $fd [lmember $slaves apu440] [lmember $masters apu440]
		set rstName reset
		set clkName apu_clk
	}
	if {[lmember $slaves apu]} {
		if {!$first} {
			puts $fd ";"
		} else {
			set first 0
		}
		puts $fd "    -- apu ports"
		genApuPorts $fd [lmember $slaves apu] [lmember $masters apu]
		set rstName reset
		set clkName apu_clk
	}
	if {[lmember $slaves pcmcia] || [lmember $masters pcmcia]} {
		if {!$first} {
			puts $fd ";"
		} else {
			set first 0
		}
		puts $fd "    -- pcmcia ports"
		genPcmciaPorts $fd [lmember $slaves pcmcia] [lmember $masters pcmcia]
		set rstName PCMCIA2IP_Reset
		set clkName PCMCIA_Clk
		#gme - fix clocking
	}
	if {[lmember $slaves opb] || [lmember $masters opb]} {
		if {!$first} {
			puts $fd ";"
		} else {
			set first 0
		}
		puts $fd "    -- opb ports"
		genOpbPorts $fd [lmember $slaves opb] [lmember $masters opb]
		set rstName Opb2IP_Reset
		set clkName Opb2IP_Clk
	}
	if {[lmember $slaves plb] || [lmember $masters plb]} {
		if {!$first} {
			puts $fd ";"
		} else {
			set first 0
		}
		puts $fd "    -- plb ports"
		genPlbPorts $fd [lmember $slaves plb] [lmember $masters plb] 0
		set rstName PLB_Rst
		set clkName PLB_Clk
	} elseif {[lmember $slaves plb4] || [lmember $masters plb4]} {
		if {!$first} {
			puts $fd ";"
		} else {
			set first 0
		}
		puts $fd "    -- plb ports"
		genPlbPorts $fd [lmember $slaves plb4] [lmember $masters plb4] 1
		set rstName PLB_Rst
		set clkName PLB_Clk
	}
	
	# Generate ports for each fsl slave connection (i.e. fsl is one bus/connection)
	set fslNo 0
	foreach con $connections {
		if {[lindex $con 0] != "memory"} {
			if {[selectBus $con] == "fsl"} {
				if {!$first} {
					puts $fd ";"
				} else {
					set first 0
				}
				puts $fd "    -- fsl$fslNo ports"
				genFslPorts $fd $fslNo [lindex $con 1]
				set rstName FSL_Rst
				set clkName FSL_Clk
				incr fslNo
			}
		}
	}

	# Generate ports for each dvi connection (i.e. dvi is one bus/connection)
	set dviNo 0
	foreach con $connections {
		if {[lindex $con 0] != "memory"} {
			if {[selectBus $con] == "dvi"} {
				if {!$first} {
					puts $fd ";"
				} else {
					set first 0
				}
				puts $fd "    -- dvi$dviNo ports"
				genDVIPorts $fd $dviNo [lindex $con 1]
				set rstName DVI_Rst
				set clkName DVI_Clk
				incr dviNo
			}
		}
	}

	foreach con $connections {
		switch [selectBus $con] {
			port    {
				if {!$first} {
					puts $fd ";"
				} else {
					set first 0
				}
				puts $fd "    -- [lindex $con 2] ports"
				genPortPorts $fd $con
			}
		}
	}

	if {$rstName == ""} {
		if {!$first} {
			puts $fd ";"
		}
		set rstName rst
		set clkName clk
		puts $fd "    rst : in std_logic;"
		puts -nonewline $fd "    clk : in std_logic"
	}

	set sclkName $clkName
	if {$::dual_clock} {
		set clkName "co_clk"
		puts -nonewline $fd ";\n    ${clkName} : in std_logic"
	}
	
	puts $fd "\n    );"
	puts $fd "end ${topname};\n"

	puts $fd "architecture impl of ${topname} is"

	generateDesignComponent $fd $name $connections

	generateDesignSignals $fd $connections

	if { $_flags(active_low_reset) } {
		# Every bus reset assumed to be active-high, so we have to invert the
		# incoming reset if the user says it's active-low instead.  We need a 
		# signal to put the inverted reset on.  The "_n" indicates the fact 
		# that the signal is an inverse, *not* its actual polarity. - rkb
		gen_signal $fd "${rstName}_n" [scalar_type]
	}

	puts $fd ""    
	foreach con $connections {
		switch [selectBus $con] {
			apu440    { genApu440ConnSignals $fd $con }
			apu    { genApuConnSignals $fd $con }
			pcmcia { genPcmciaConnSignals $fd $con }
			opb    { genOpbConnSignals $fd $con }
			plb    { genPlbConnSignals $fd $con }
			plb4    { genPlbConnSignals $fd $con }
		}
	}

	foreach bus $slaves {
		puts $fd "\n  -- $bus signals"
		switch $bus {
			apu440    { genApu440Signals $fd $attachments(apu440) }
			apu    { genApuSignals $fd $attachments(apu) }
			pcmcia { genPcmciaSignals $fd $attachments(pcmcia) }
			opb    { genOpbSignals $fd $attachments(opb) }
			plb    { genPlbSignals $fd $attachments(plb) 0 }
			plb4    { genPlbSignals $fd $attachments(plb4) 1 }
		}
		puts $fd "  -- end $bus signals"
	}

	puts $fd "begin"

	# The Impulse C module will invert its own reset if the active-low option
	# is on (to support an active-low reset in modules not on a bus).
	set icModuleReset $rstName
	if { $_flags(active_low_reset) } {
		if { $rstName != "FSL_Rst"} {
			puts "Warning: Active-low reset not supported in the bus wrapper.  $rstName must be connected to an active-high signal."
		}
		gen_signal_assignment $fd "${rstName}_n" [negate $rstName]

		set rstName "${rstName}_n"
		# Sub-components of this wrapper that need an active-high reset 
		# (stream_to_BUSNAME, etc.) *should* use $rstName--now the inverse of 
		# the bus reset.  Only FSL actually does that, though. - rkb
	}

	foreach bus $slaves {
		switch $bus {
			apu440    { genApu440Ipif $fd $multibus $attachments(apu440) }
			apu    { genApuIpif $fd $multibus $attachments(apu) }
			pcmcia { genPcmciaIpif $fd $multibus $attachments(pcmcia) }
			opb    { genOpbIpif $fd $multibus }
			plb    { genPlbIpif $fd $multibus $attachments(plb) 0 }
			plb4    { genPlbIpif $fd $multibus $attachments(plb4) 1 }
		}
	}

	set connNo(apu440)    	0
	set udiNo(apu440)    	0
	set connNo(apu)    	0
	set connNo(pcmcia) 	0
	set connNo(opb)    	0
	set connNo(plb)    	0
	set connNo(plb4)    	0
	set connNo(fsl)    	0
	set connNo(dvi)    	0
	set connNo(port)    	0
	set connNo(generic)	0
	foreach con $connections {
		set busname [selectBus $con]
		# Note that only FSL actually uses the $rstName and $sclkName
		# arguments. - rkb
		switch $busname {
		    apu440 {
			if {[lindex $con 0] != "instruction"} {
				genApu440Conn $fd $con $connNo(apu440) $attachments(apu440) $rstName $sclkName 
			} else {
				genApu440Conn $fd $con $udiNo(apu440) $attachments(apu440) $rstName $sclkName 
			}
		    }
		    apu    { genApuConn $fd $con $connNo(apu) $attachments(apu) $rstName $sclkName }
		    pcmcia { genPcmciaConn $fd $con $connNo(pcmcia) $attachments(pcmcia) $rstName $sclkName }
		    opb    { genOpbConn $fd $con $connNo(opb) $attachments(opb) $rstName $sclkName }
		    plb    { genPlbConn $fd $con $connNo(plb) $attachments(plb) $rstName $sclkName 0 }
		    plb4    { genPlbConn $fd $con $connNo(plb4) $attachments(plb4) $rstName $sclkName 1 }
		    fsl    { genFslConn $fd $con $connNo(fsl) $rstName $sclkName}
		    dvi    { genDVIConn $fd $con $connNo(dvi) $rstName $sclkName}
		    port   { genPortConn $fd $con }
		}
		if {[lindex $con 0] != "memory"} {
		    if {[lindex $con 0] != "instruction"} {
			incr connNo($busname)
		    } else {
			incr udiNo($busname)
		    }
		}
	}
    
	if {[lmember $slaves apu440]} {
		genApu440OutputSignals $fd $connections
	}
	if {[lmember $slaves apu]} {
		genApuOutputSignals $fd $connections
	}
	if {[lmember $slaves pcmcia]} {
		set selWidth [calcAddrBits $attachments(pcmcia)]
		set addrWidth [expr $selWidth + 4]
		if {$addrWidth > 0} {
			puts $fd "  pcmciasel <= PCMCIA2IP_Addr([expr $addrWidth-1] downto 4);\n"
		}
		genOutputMux $fd pcmcia pcmciasel $selWidth IP2PCMCIA_Data $connections
	}
   
	if {[lmember $slaves opb]} {
		set addrWidth [calcAddrBits $attachments(opb)]
		if {$addrWidth > 0} {
			puts $fd "  opbsel <= Opb2IP_Addr([expr 28 - $addrWidth] to 27);\n"
		}
		genOutputMux $fd opb opbsel $addrWidth IP2Opb_Data $connections
	}

	if {[lmember $slaves plb]} {
		set selWidth [calcAddrBits $attachments(plb)]
		set addrWidth [expr $selWidth + 4]
		if {$addrWidth > 0} {
			puts $fd "  plbsel <= Plb2IP_raddr([expr $addrWidth - 1] downto 4);\n"
		}
		genOutputMux $fd plb plbsel $selWidth IP2Plb_Data $connections
	} elseif {[lmember $slaves plb4]} {
		set selWidth [calcAddrBits $attachments(plb4)]
		set addrWidth [expr $selWidth + 4]
		if {$addrWidth > 0} {
			puts $fd "  plbsel <= Plb2IP_raddr([expr $addrWidth - 1] downto 4);\n"
		}
		genOutputMux $fd plb4 plbsel $selWidth IP2Plb_Data $connections
	}

	generateDesignInstance $fd $name $icModuleReset $sclkName $clkName $connections

	puts $fd "end;\n"

	close $fd
}

# Generate a Tcl script used by EDK to create a user constraints file.  The 
# script prints the contents of user .ucf files to create the pcore's .ucf.
proc GeneratePcoreTcl {outdir name} {
	set outfile [file join $outdir "pcore/${name}"]
	if {[catch {set fd [open $outfile w]}]} {
		puts "unable to create $outfile"
		exit 1
	}

	puts $fd "proc generate_corelevel_ucf \{mhsinst\} \{"
	puts $fd "  set filePath \[xget_ncf_dir \$mhsinst\]"
	puts $fd "  file mkdir    \$filePath"
	puts $fd "  set    instname   \[xget_hw_parameter_value \$mhsinst \"INSTANCE\"\]"
	puts $fd "  set    ipname     \[xget_hw_option_value    \$mhsinst \"IPNAME\"\]"
	puts $fd "  set    name_lower \[string 	  tolower    \$instname\]"
	puts $fd "  set    fileName   \$name_lower"
	puts $fd "  append fileName   \"_wrapper.ucf\""
	puts $fd "  append filePath   \$fileName"
	puts $fd "  if \{\[catch \{set fd \[open \$filePath w\]\}\]\} \{"
	puts $fd "    puts \"Unable to create \$filePath\""

	puts $fd "    return 1"
	puts $fd "  \}"
	foreach fname $::hdlfiles {
		if {[file extension $fname] == ".ucf"} {
			if {![catch {set ucfd [open [file join $outdir $fname] r]}]} {
				while {-1 != [gets $ucfd line]} {
					puts $fd "  puts \$fd \{$line\}"
				}
				close $ucfd
			}
		}
	}
	puts $fd "  close \$fd"
	puts $fd "  puts   \[xget_ncf_loc_info \$mhsinst\]"
	puts $fd "\}"
}
		
proc GenerateMPD {outdir topname name connections slaves masters} {
	set outfile [file join $outdir "pcore/${topname}_v2_1_0.mpd"]
	if {[catch {set fd [open $outfile w]}]} {
		puts "unable to create $outfile"
		exit 1
	}
    
	puts $fd "BEGIN ${topname}\n"
 	puts $fd "OPTION IPTYPE = PERIPHERAL"
	if {[lmember $slaves apu440]} {
		puts $fd "OPTION ARCH_SUPPORT_MAP = (virtex5fx=DEVELOPMENT)"
	} elseif {[lmember $slaves apu]} {
		puts $fd "OPTION ARCH_SUPPORT_MAP = (virtex4fx=DEVELOPMENT)"
	} else {
		puts $fd "OPTION ARCH_SUPPORT_MAP = (others=DEVELOPMENT)"
	}

#if { $::libCoreGenCommon::NumCoreGens } {
	puts $fd "OPTION STYLE = MIX"
#} else {
#	puts $fd "OPTION STYLE = HDL"
#}
 
 	# Determine if this pcore is mixed-language.  Look for Verilog files in the
 	# user-supplied IP.
 	set isMixedHDL 0
	# Determine if this pcore contains user constraints
	set hasUCF 0
 	foreach fname $::hdlfiles {
 		if {[file extension $fname] == ".v"} {
 			set isMixedHDL 1
 		}
 		if {[file extension $fname] == ".ucf"} {
 			set hasUCF 1
 		}
 	}
 	if {$isMixedHDL} {
 		puts $fd "OPTION HDL = MIXED"
 	} else {
 		puts $fd "OPTION HDL = VHDL"
 	}
	if {$hasUCF} {
		if { ! [lmember $slaves apu440] && ! [lmember $slaves apu]} {
			# APU pcores already set these options
			puts $fd "OPTION RUN_NGCBUILD = TRUE"
			puts $fd "OPTION IMP_NETLIST = TRUE"
		}
		puts $fd "OPTION PLATGEN_SYSLEVEL_UPDATE_PROC = generate_corelevel_ucf"
		GeneratePcoreTcl $outdir "${topname}_v2_1_0.tcl"
	}
	
	# set callback for elaborating any XCOs exported OPTION ELABORATE_PROC = impulse_elaborate
	puts $fd "OPTION ELABORATE_PROC = impulse_elaborate"
 
 	puts $fd ""

	if {[lmember $slaves opb]} {
		puts $fd "BUS_INTERFACE BUS = SOPB, BUS_STD = OPB, BUS_TYPE = SLAVE"
	}
	if {[lmember $masters opb]} {
		puts $fd "BUS_INTERFACE BUS = MOPB, BUS_STD = OPB, BUS_TYPE = MASTER"
	}
	if {[lmember $slaves plb]} {
		puts $fd "BUS_INTERFACE BUS = SPLB, BUS_STD = PLB, BUS_TYPE = SLAVE"
	} elseif {[lmember $slaves plb4]} {
		puts $fd "BUS_INTERFACE BUS = SPLB, BUS_STD = PLBV46, BUS_TYPE = SLAVE"
	}
	if {[lmember $masters plb]} {
		puts $fd "BUS_INTERFACE BUS = MPLB, BUS_STD = PLB, BUS_TYPE = MASTER"
	} elseif {[lmember $masters plb4]} {
		puts $fd "BUS_INTERFACE BUS = MPLB, BUS_STD = PLBV46, BUS_TYPE = MASTER"
	}
	set no 0
	foreach con $connections {
		set busname [selectBus $con]
		if {$busname == "fsl"} {
			set dir [lindex $con 1]
			if {$dir == "in"} {
				puts $fd "BUS_INTERFACE BUS = SFSL$no, BUS_STD = FSL, BUS_TYPE = SLAVE"
			} else {
				puts $fd "BUS_INTERFACE BUS = MFSL$no, BUS_STD = FSL, BUS_TYPE = MASTER"
			}
			incr no
		}
		if {$busname == "dvi"} {
			set dir [lindex $con 1]
			if {$dir == "in"} {
				puts $fd "BUS_INTERFACE BUS = IMPULSE_DVI_IN, BUS_STD = DVI_VIDEO, BUS_TYPE = TARGET"
			} else {
				puts $fd "BUS_INTERFACE BUS = IMPULSE_DVI_OUT, BUS_STD = DVI_VIDEO, BUS_TYPE = INITIATOR"
			}
			incr no
		}
		if {$busname == "port"} {
			set bus [lindex $con 4]
			if {$bus == "bus_signals"} {
				set conType [lindex $con 0]
				set bus_std [string toupper "co_${conType}"]
				set dir [lindex $con 1]
				set con_name [string toupper [lindex $con 2]]
				#puts "bus_signals:$bus_std"
				# co_register is pt-to-pt for now, master/slave bus requires full bus 
				#  implementation within EDK
#				if { $conType == "register" } {
#					# one-to-many
#					if {$dir == "in"} {
#						puts $fd "BUS_INTERFACE BUS = $con_name, BUS_STD = $bus_std, BUS_TYPE = SLAVE"
#					} else {
#						puts $fd "BUS_INTERFACE BUS = $con_name, BUS_STD = $bus_std, BUS_TYPE = MASTER"
#					}
#				} else {
					# point-to-point
					if {$dir == "in"} {
						puts $fd "BUS_INTERFACE BUS = $con_name, BUS_STD = $bus_std, BUS_TYPE = TARGET"
					} else {
						puts $fd "BUS_INTERFACE BUS = $con_name, BUS_STD = $bus_std, BUS_TYPE = INITIATOR"
					}
#				}
			}
		}
	}
	puts $fd ""

	if {[lmember $slaves apu440]} {
		puts $fd "PARAMETER c_family          = virtex5fx,    DT = string\n"
	} elseif {[lmember $slaves apu] || [lmember $slaves pcmcia]} {
		puts $fd "PARAMETER c_family          = virtex4,    DT = string\n"
	} else {
		puts $fd "PARAMETER c_family          = virtex2,    DT = string\n"
	}

	if {[lmember $slaves plb] && [lmember $slaves opb]} {
		set multibus 1
	} elseif {[lmember $slaves plb4] && [lmember $slaves opb]} {
		set multibus 1
	} else {
		set multibus 0
	}
	set noclock 1
	if {[lmember $slaves apu440]} {
		genApu440MPD $fd
		set noclock 0
	}
	if {[lmember $slaves apu]} {
		genApuMPD $fd
		set noclock 0
	}
	if {[lmember $slaves opb] || [lmember $masters opb]} {
		genOpbMPD $fd $multibus [lmember $slaves opb] [lmember $masters opb]
		set noclock 0
	}
	if {[lmember $slaves plb] || [lmember $masters plb]} {
		genPlbMPD $fd $multibus [lmember $slaves plb] [lmember $masters plb] 0
		set noclock 0
	} elseif {[lmember $slaves plb4] || [lmember $masters plb4]} {
		genPlbMPD $fd $multibus [lmember $slaves plb4] [lmember $masters plb4] 1
		set noclock 0
	}
	if {[lmember $slaves fsl]} {
		genFslMPD $fd $connections
		set noclock 0
	}
	if {[lmember $slaves dvi]} {
		genDVIMPD $fd $connections
		set noclock 0
	}
	if {[lmember $slaves port]} {
		genPortMPD $fd $connections
	}
 	if {$noclock} {
 		puts $fd "PORT clk = \"\", DIR=IN, SIGIS=CLK"
 		puts $fd "PORT rst = \"\", DIR=IN, SIGIS=RST"
	}
	if {$::dual_clock} {
		puts $fd "PORT co_clk = \"\", DIR = IN, SIGIS=CLK"
	}

	puts $fd "END"

	close $fd
}

proc GeneratePAO {outdir topname name connections slaves masters} {
	set outfile [file join $outdir "pcore/${topname}_v2_1_0.pao"]
	if {[catch {set fd [open $outfile w]}]} {
		puts "unable to create $outfile"
		exit 1
	}
    
	if {[lmember $slaves opb]} {
		puts $fd "lib proc_common_v1_00_b proc_common_pkg vhdl"
		puts $fd "lib proc_common_v1_00_b pselect vhdl"
		puts $fd "lib proc_common_v1_00_b or_muxcy vhdl\n"

		puts $fd "lib ipif_common_v1_00_c ipif_pkg vhdl"
		puts $fd "lib ipif_common_v1_00_c interrupt_control vhdl"
		puts $fd "lib ipif_common_v1_00_c ipif_steer vhdl\n"

		puts $fd "lib opb_ipif_v3_00_a    reset_mir vhdl"
		puts $fd "lib opb_ipif_v3_00_a    opb_bam vhdl"
		puts $fd "lib opb_ipif_v3_00_a    opb_ipif vhdl\n"

		puts $fd "lib opb_ipif_ssp0_v1_00_b  opb_ipif_ssp0 vhdl\n"
	}

        if {[lsearch [getLibraryList] "float"] >=0 } {
	        puts $fd "lib impulse float_ll vhdl"
	}
        if {[lsearch [getLibraryList] "float_fast"] >=0 } {
	        puts $fd "lib impulse float_fast vhdl"
	}
	puts $fd "lib impulse divmod vhdl"
	puts $fd "lib impulse gmem vhdl"
	puts $fd "lib impulse fifo vhdl"
	puts $fd "lib impulse fifo_dc vhdl"
	puts $fd "lib impulse csignal vhdl"
	puts $fd "lib impulse cregister vhdl"
	puts $fd "lib impulse stream vhdl"
	puts $fd "lib impulse stream_dc vhdl"
	puts $fd "lib impulse impack vhdl"
	puts $fd "lib impulse mem_if vhdl"

	if {[lmember $slaves apu440]} {
		puts $fd "lib impulse apu440_if vhdl"
	}
	if {[lmember $slaves apu]} {
		puts $fd "lib impulse apu_if vhdl"
	}
	if {[lmember $slaves fsl]} {
		puts $fd "lib impulse fsl_if vhdl"
	}
	if {[lmember $slaves dvi]} {
		puts $fd "lib impulse dvi_if vhdl"
	}
	if {[lmember $slaves opb] || [lmember $masters opb]} {
		puts $fd "lib impulse opb_if vhdl"
	}
	if {[lmember $slaves plb] || [lmember $masters plb]} {
		puts $fd "lib impulse plbdma vhdl"
		puts $fd "lib impulse plb_if vhdl"
	} elseif {[lmember $slaves plb4] || [lmember $masters plb4]} {
		puts $fd "lib impulse plbdma4 vhdl"
		puts $fd "lib impulse plb4_if vhdl"
	}

	# include files for CoreGens in library impulse
#	global env
#	# need to add each hdl file in each libray's 'hdl' dir
#	set coreGenLibBase "$env(IMPULSEC_HOME)/Architectures/VHDL/Xilinx/CoreGen_libs/"
#	foreach libName $::libCoreGenCommon::CoreGenLibList {
#		puts "$libName is a CoreGen library"
#		foreach libHdlFile [glob -nocomplain $coreGenLibBase/$libName/hdl/*.vhd] {
#			puts $fd "lib impulse [file tail $libHdlFile] vhdl"
#		}
#	}
#	# need to add each hdl file from coregen directory
#	foreach libHdlFile [glob -nocomplain $outdir/coregen/*.vhd] {
#		puts $fd "lib impulse [file tail $libHdlFile] vhdl"
#	}

	foreach fname $::hdlfiles {
 		if {[file extension $fname] == ".v"} {
 			set hdllang "verilog"
 		} else {
 			set hdllang "vhdl"
 		}
		if {[file extension $fname] != ".ucf"} {
			puts $fd "lib ${topname}_v1_00_a [file rootname $fname] ${hdllang}"
		}
	}
	puts $fd "lib ${topname}_v1_00_a $topname vhdl"

	close $fd
}

proc GenerateTcl {outdir topname name connections slaves masters} {
	# nothing currently to gnerate, just need to copy a few tcl files 
	global env
	set autoipCoregenTclDir "$env(IMPULSEC_HOME)/Architectures/Xilinx/AutoIP/Common/COREGenerator"
	set autoipTclFilesToCopy {"coregen.tcl" "xco.tcl" }
	
	set outdirTcl [file join $outdir "tcl"]
	file mkdir $outdirTcl
	
	foreach autoipTclF $autoipTclFilesToCopy {
		file copy -force [file join $autoipCoregenTclDir $autoipTclF] [file join $outdirTcl $autoipTclF]
	}
	# copy PLB pcore script
	set pcoreTclDir "$env(IMPULSEC_HOME)/Architectures/VHDL/Xilinx"
	file copy -force [file join $pcoreTclDir "pcore_v2_1_0.tcl"] [file join $outdirTcl "${topname}_v2_1_0.tcl"]
}

proc GenerateBBD {outdir topname name connections slaves masters} {
	# CORE Generator from CoDeveloper is being removed, just create
	#  dummy BBD file and necessary empty NGC file to keep EDK happy
	#  until BBD is over-written during elaboration
	set outdirPcore [file join $outdir "pcore"]

	set outdirNetlist [file join $outdir "netlist"]
	file mkdir $outdirNetlist
	
	set bbdFile [open [file join $outdirPcore "${topname}_v2_1_0.bbd"] w]
	puts $bbdFile "# THIS IS INTENTIONALLY LEFT EMPTY AND WILL BE OVERWRITTEN DURING EDK'S ELABORATION STAGE FOR THIS PCORE"
	puts $bbdFile "FILES"
	puts $bbdFile "dummy.ngc"
	close $bbdFile

	set dmyNgcFile [open [file join $outdirNetlist "dummy.ngc"] w]
	puts $dmyNgcFile "THIS IS A DUMMY NGC FILE TO PLACATE EDK WHEN SCANNING THE USER REPOSITORY"
	close $dmyNgcFile
}

proc GenerateBUS {outdir name connections} {
	set slaves [list]
	set masters [list]

	#####
	# NOTE: Calling CORE Generator from CoDeveloper is being deprecated with move
	#        to ISE (which will call it automatically) and including a Tcl script
	#        for EDK - this was necessary to migrate to Xilinx FP v5 cores using 
	#		   autoip feature and the handling of XCO files is the model to use 
	#			for future CORE Generator IP 
	#####
	# load and run libCoreGenCommon
	#global env
	#source "$env(IMPULSEC_HOME)/Architectures/VHDL/Xilinx/CoreGen_libs/common/libCoreGenCommon.tcl"
	#libCoreGenCommon::GenerateCoreGen $outdir $name $connections

	foreach con $connections {
		set busname [selectBus $con]
		if {[lindex $con 0] != "memory"} {
			if {[lsearch $slaves $busname] < 0} {
				lappend slaves $busname
			}
		} else {
			if {[lsearch $masters $busname] < 0} {
				lappend masters $busname
			}
		}
	    if {[lindex $con 0] == "instruction"} {
		checkInstruction $con
	    }
	}

	# determine pcore prefix by priority if multiple busses
	if {[lmember $slaves apu440]} {
	        genApu440Init
		set topname apu_$name
	} elseif {[lmember $slaves apu]} {
		set topname apu_$name
	} elseif {[lmember $slaves pcmcia]} {
		set topname pcmcia_$name
	} elseif {[lmember $slaves fsl]} {
		set topname fsl_$name
	} elseif {[lmember $slaves plb] || [lmember $masters plb]} {
		set topname plb_$name
	} elseif {[lmember $slaves plb4] || [lmember $masters plb4]} {
		set topname plb_$name
	} elseif {[lmember $slaves dvi]} {
		set topname dvi_$name
	} elseif {[lmember $slaves opb] || [lmember $masters opb]} {
		set topname opb_$name
	} elseif {[lmember $slaves port]} {
		set topname impc_$name
	} else {
		# No need to generate bus mux or pcore if we're not using one
		# of the supported buses.
		return
	}

	GenerateTop $outdir $topname $name $connections $slaves $masters
	GenerateMPD $outdir $topname $name $connections $slaves $masters

	GeneratePAO $outdir $topname $name $connections $slaves $masters
	GenerateTcl $outdir $topname $name $connections $slaves $masters
	GenerateBBD $outdir $topname $name $connections $slaves $masters
	
	# Got here, assuming EDK is target, call into libCoreGenCommon to generate
	#  files for inclusion of CoreGen into pcore
	#libCoreGenCommon::GenerateCoreGenFilesForEDK $outdir $topname $connections
}
