--------------------------------------------------------------------------
-- Copyright (c) 2002-2007 by Impulse Accelerated Technologies, Inc.    --
-- All rights reserved.                                                 --
--                                                                      --
-- This source file may be used and redistributed without charge        --
-- subject to the provisions of the IMPULSE ACCELERATED TECHNOLOGIES,   --
-- INC. REDISTRIBUTABLE IP LICENSE AGREEMENT, and provided that this    --
-- copyright statement is not removed from the file, and that any       --
-- derivative work contains this copyright notice.                      - 
--------------------------------------------------------------------------

-- Generate an Impulse C acknowledge signal from floating_point core's rdy
-- signal.  The rdy signal is a single cycle pulse, but Impulse C requires that
-- the acknowledge be held active.

library ieee;
use ieee.std_logic_1164.all;
entity impulse_acknowledge is 
  port (
    clk : in  std_ulogic;
    req : in  std_ulogic;
    rdy : in  std_ulogic;
    ack : out std_ulogic);
end impulse_acknowledge;

architecture rtl of impulse_acknowledge is
  signal complete : std_ulogic;
begin
  genack: process (clk)
  begin  -- process genack
    if clk'event and clk = '1' then  -- rising clock edge
      if (req = '1') then
        complete <= '0'; -- new request, clear the complete status
      elsif (rdy = '1') then
        complete <= '1'; -- got result, set the complete status
      end if;
    end if;
  end process genack;

  ack <= complete;
end rtl;

library ieee;
use ieee.std_logic_1164.all;

entity fneg_ll is
  port (
    a: in std_ulogic_vector(31 downto 0);
    result: out std_ulogic_vector(31 downto 0));
end fneg_ll;

architecture fneg_ll_a of fneg_ll is
begin
  result(31) <= a(31) xor '1';
  result(30 downto 0) <= a(30 downto 0);
end fneg_ll_a;

library ieee;
use ieee.std_logic_1164.all;

entity fabs_ll is
  port (
    a: in std_ulogic_vector(31 downto 0);
    result: out std_ulogic_vector(31 downto 0));
end fabs_ll;

architecture fabs_ll_a of fabs_ll is
begin
  result(31) <= '0';
  result(30 downto 0) <= a(30 downto 0);
end fabs_ll_a;

library ieee;
use ieee.std_logic_1164.all;

entity fnegd_ll is
  port (
    a: in std_ulogic_vector(63 downto 0);
    result: out std_ulogic_vector(63 downto 0));
end fnegd_ll;

architecture fnegd_ll_a of fnegd_ll is
begin
  result(63) <= a(63) xor '1';
  result(62 downto 0) <= a(62 downto 0);
end fnegd_ll_a;

library ieee;
use ieee.std_logic_1164.all;

entity fabsd_ll is
  port (
    a: in std_ulogic_vector(63 downto 0);
    result: out std_ulogic_vector(63 downto 0));
end fabsd_ll;

architecture fabsd_ll_a of fabsd_ll is
begin
  result(63) <= '0';
  result(62 downto 0) <= a(62 downto 0);
end fabsd_ll_a;


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;

entity ftod_ll is
  port (
    clk: in std_ulogic;
    a: in std_ulogic_vector(31 downto 0);
    result: out std_ulogic_vector(63 downto 0));
end ftod_ll;

architecture ftod_ll_a of ftod_ll is
    signal exp_s : std_ulogic_vector(7 downto 0);
    signal exp_d : unsigned(10 downto 0);
    signal zero : std_ulogic;
begin
  exp_s <= a(30 downto 23);             -- extract the exponent
  exp_d <= unsigned("000" & exp_s) + 896;  -- rebias
  zero <= '1' when a = X"00000000" else '0';

  result(63) <= a(31);                  -- sign bit
   -- biased exponent
  result(62 downto 52) <= std_ulogic_vector(exp_d) when zero = '0'
                          else "00000000000";
  result(51 downto 29) <= a(22 downto 0);  -- fraction
  result(28 downto 0) <= "00000000000000000000000000000";
end ftod_ll_a;
  
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;

entity dtof_ll is
  port (
    clk: in std_ulogic;
    a: in std_ulogic_vector(63 downto 0);
    result: out std_ulogic_vector(31 downto 0));
end dtof_ll;

architecture dtof_ll_a of dtof_ll is
    signal exp_s : std_ulogic_vector(7 downto 0);
    signal exp_d : std_ulogic_vector(10 downto 0);
    signal exp : unsigned(10 downto 0);
    signal ovr, inf : std_ulogic;
begin
  exp_d <= a(62 downto 52);             -- extract the exponent
  exp <= unsigned(exp_d) - 896;  -- rebias
  exp_s <= std_ulogic_vector(exp(7 downto 0));
  ovr <= '0' when std_ulogic_vector(exp(10 downto 8)) = "000" else '1';
  inf <= a(62) when ovr = '1' else '0';

  -- sign bit
  result(31) <= a(63);
  -- exponent
  result(30 downto 23) <= exp_s when ovr = '0'
                          else "00000000" when inf = '0'
                          else "11111111";
  -- fraction
  result(22 downto 0) <= a(51 downto 29) when ovr = '0'
                         else "00000000000000000000000";
end dtof_ll_a;

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity latency_delay is
    generic (
        WIDTH : std_ulogic_vector(7 downto 0) := "00000001";
        LATENCY : std_ulogic_vector(3 downto 0) := "0001"
    );
    port (
        clk: in std_ulogic;
        a: in std_ulogic_vector(to_integer(unsigned(WIDTH))-1 downto 0);
        a_dly: out std_ulogic_vector(to_integer(unsigned(WIDTH))-1 downto 0);
        pipeEn: in std_ulogic
    );
end latency_delay;

architecture latency_delay_a of latency_delay is
    constant UWIDTH : integer := to_integer(unsigned(WIDTH));
    constant ULATENCY : integer := to_integer(unsigned(LATENCY));
    type a_dly_arrtype is array (0 to ULATENCY) of std_ulogic_vector(UWIDTH-1 downto 0);
    signal a_dly_arr: a_dly_arrtype;
begin
    a_dly_arr(0) <= a;
    
    gen_a_dly: for I in 1 to ULATENCY generate
        process (clk)
        begin
            if clk'event and clk = '1' then
                if (pipeEn = '1') then
	                a_dly_arr(I) <= a_dly_arr(I-1);
                end if;
            end if;
        end process;
    end generate gen_a_dly;
    
    a_dly <= a_dly_arr(ULATENCY);
end latency_delay_a;

	  
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

package xilinx_float_low_latency is

  component fneg_ll is
    port (
    a: in std_ulogic_vector(31 downto 0);
    result: out std_ulogic_vector(31 downto 0));
  end component;

  component fabs_ll is
    port (
    a: in std_ulogic_vector(31 downto 0);
    result: out std_ulogic_vector(31 downto 0));
  end component;

  component fnegd_ll is
    port (
    a: in std_ulogic_vector(63 downto 0);
    result: out std_ulogic_vector(63 downto 0));
  end component;

  component fabsd_ll is
    port (
    a: in std_ulogic_vector(63 downto 0);
    result: out std_ulogic_vector(63 downto 0));
  end component;

  component ftod_ll is
    port (
    clk: in std_ulogic;
    a: in std_ulogic_vector(31 downto 0);
    result: out std_ulogic_vector(63 downto 0));
  end component;

  component dtof_ll is
    port (
    clk: in std_ulogic;
    a: in std_ulogic_vector(63 downto 0);
    result: out std_ulogic_vector(31 downto 0));
  end component;

  component latency_delay is
    generic (
        WIDTH : std_ulogic_vector(7 downto 0) := "00000001";
        LATENCY : std_ulogic_vector(3 downto 0) := "0001"
    );
    port (
        clk: in std_ulogic;
        a: in std_ulogic_vector(to_integer(unsigned(WIDTH))-1 downto 0);
        a_dly: out std_ulogic_vector(to_integer(unsigned(WIDTH))-1 downto 0);
        pipeEn: in std_ulogic
    );
  end component;


end xilinx_float_low_latency;
