#####
# XCO file manipulation routines
#####
namespace eval ::xilinx::xcosim {

	# Parse and replace portions of xcoSimFileName writing result to fd
	# Input:
	# 	xcoSimFileName=a copy of the <coreName>.vhd output from CORE Generator
	# 	 wherein lies a simulation wrapper for the core.  This file is parsed to replace
	#	 names to follow paramArrIn(component_name) and other XCO paramaters that appear
	#	 as-is as generics.  All comments in the xcoSimFileName file are removed.
	#	paramArrIn=generic paramter array used to set generics for component
	#	portsList=ports list used to ensure ports and component signal widths match and are correct type
	#		array <port name>'s to lists:{<0=in/1=out> <width 0=signal, n>0=vector> [<signal type>]}
	# Output:
	#	Simulation definition surrounded by "-- syntehsis off/on" statments
	proc parseAndWriteXcoSimFile { fd xcoSimFileName paramArrIn portsArrIn} {
		upvar $paramArrIn paramArr
		upvar $portsArrIn portsArr
		puts $fd "-- Simulation wrapper adapted from file:$xcoSimFileName"
		puts $fd "-- Wrapper makes use of XilinxCoreLib and provided for HDL simulation" 
		puts $fd "--  without having to run CORE Generator first"
		puts $fd "-- synthesis translate_off"
		
		set fin [open $xcoSimFileName r]
		set enitiyName "unknown"
		set archModelName "unknown"
		set wrapperCompName "unknown"
		while {[gets $fin line] >= 0} {
			# remove any comments
			set comPos [string first "--" $line]
			if {$comPos>= 0} {
				set line [string trim [string range $line 0 [expr $comPos - 1]]]
			}
			if {[string length $line] > 0} {
				if {[regexp -nocase {entity (.*?) is} $line whole name]} {
						# looking for entity declaration "entity <name> is"
						set line "entity $paramArr(component_name) is"
						set enitiyName $name
				} elseif {[regexp -nocase "end $enitiyName;" $line]} {
        				# looking for entity declaration "end <name>;"
						set line "end $paramArr(component_name);"
				} elseif {[regexp -nocase "architecture (.*?) of $enitiyName is" $line whole model]} {
        				# looking for architecture definition "architecture <model> of <name> is"
    					set line "architecture $paramArr(component_name)_sim of $paramArr(component_name) is"
    					set archModelName $model
				} elseif {[regexp -nocase {component (.*$)} $line whole name]} {
        				# looking for wrapper component  declaration "component <comp_name>"
    					set line "component $paramArr(component_name)_sim_wrapper"
    					set wrapperCompName $name
				} elseif {[regexp -nocase "(.*?)for all : $wrapperCompName (.*$)" $line whole begin end]} {
        				# looking for wrapper "for all" statement "for all : <comp_name> ..."
						set line "$begin for all : $paramArr(component_name)_sim_wrapper $end"
				} elseif {[regexp -nocase "(\w*.*?) : $wrapperCompName" $line whole begin]} {
        				# looking for instance "<inst name> : <comp_name>"
						set line "$begin : $paramArr(component_name)_sim_wrapper"
				} elseif {[regexp -nocase "end $archModelName;" $line]} {
        				# looking for end of architecture definition "end <model>;"
						set line "end $paramArr(component_name)_sim;"
				} elseif {[regexp -nocase {(^.*?)(c_.*?)=>(.+?)([,;\)]*$)} $line whole begin generic val end]} {
					# looking for generic setting "c_<name> => <value>"
					set generic [string trim $generic]
					if {[info exists paramArr($generic)]} {
						set line "$begin${generic} => $paramArr($generic)$end"
					}
				} elseif {[regexp -nocase {(\s*)(\w*?):\s*(in|out)(.+?)(\)\);|\);|;)$} $line whole begin name dir sigtype end]} {
					# looking for ports or component ports: <port name>:<direction> <signal type>
#					puts "FOUND PORT:$line"
#					puts "--\[$begin\]\[$name\]\[$dir\]\[$sigtype\]\[$end\]"
					if {[info exists portsArr($name)]} {
						set portDef $portsArr($name)
#						puts "portDef=$portDef"
						set line "$begin${name} : "
						# direction
						if {[lindex $portDef 0] == 1} {
							append line "out "
						} else {
							append line "in "
						}
						# signal type and size - assumes type is specified in array data
						set portLen [lindex $portDef 1] 
						set portType [lindex $portDef 2]
						if {[lindex $portDef 1] == 0} {
							append line "$portType"
						} else {
							set msb [expr $portLen - 1]
							append line "${portType}(${msb} downto 0"
						}
						append line $end
#						puts "NEW PORT:$line"
					}
				}				
				puts $fd $line
			}
		}
		close $fin
		puts $fd "-- synthesis translate_on"
		puts $fd ""
	}
}

return 0